"""
Copyright (c) 2025, Auorui.
All rights reserved.

Swin Transformer: Hierarchical Vision Transformer using Shifted Windows
    <https://arxiv.org/pdf/2103.14030>
use for reference: https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/swin_transformer.py
                   https://github.com/WZMIAOMIAO/deep-learning-for-image-processing/blob/master/pytorch_classification/swin_transformer/model.py
"""
from functools import partial

import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from pyzjr.utils.converter import to_2tuple
from pyzjr.nn.models.bricks.drop import DropPath
from pyzjr.nn.models.bricks.initer import trunc_normal_

LayerNorm = partial(nn.LayerNorm, eps=1e-6)

class PatchPartition(nn.Module):
    def __init__(self, patch_size=4, in_channels=3, embed_dim=96, norm_layer=None):
        super().__init__()
        self.patch_size = to_2tuple(patch_size)
        self.embed_dim = embed_dim
        self.proj = nn.Conv2d(in_channels, self.embed_dim,
                              kernel_size=self.patch_size, stride=self.patch_size)
        self.norm = norm_layer(self.embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        _, _, H, W = x.shape
        if H % self.patch_size[0] != 0:
            pad_h = self.patch_size[0] - H % self.patch_size[0]
            x = F.pad(x, (0, 0, 0, pad_h))
        if W % self.patch_size[1] != 0:
            pad_w = self.patch_size[1] - W % self.patch_size[1]
            x = F.pad(x, (0, pad_w, 0, 0))
        x = self.proj(x)     # [B, embed_dim, H/patch_size, W/patch_size]
        Wh, Ww = x.shape[2:]
        x = x.flatten(2).transpose(1, 2)    # [B, num_patches, embed_dim]
        # Linear Embedding
        x = self.norm(x)
        # x = x.transpose(1, 2).view(-1, self.embed_dim, Wh, Ww)
        return x, Wh, Ww

class MLP(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop_ratio=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop_ratio)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x

class PatchMerging(nn.Module):
    def __init__(self, dim, norm_layer=LayerNorm):
        super().__init__()
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x, H, W):
        """
        Args:
            x: Input feature, tensor size (B, H*W, C).
            H, W: Spatial resolution of the input feature.
        """
        B, L, C = x.shape
        assert L == H * W, "input feature has wrong size"

        x = x.view(B, H, W, C)

        if H % 2 == 1 or W % 2 == 1:
            x = F.pad(x, (0, 0, 0, W % 2, 0, H % 2))

        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
        x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
        x = x.view(B, -1, 4 * C)  # B H/2*W/2 4*C

        x = self.norm(x)
        x = self.reduction(x)

        return x

class WindowAttention(nn.Module):
    """
    Window based multi-head self attention (W-MSA) module with relative position bias.
    It supports shifted and non-shifted windows.
    """
    def __init__(
            self,
            dim,
            window_size,
            num_heads,
            qkv_bias=True,
            proj_bias=True,
            attention_dropout_ratio=0.,
            proj_drop=0.,
    ):
        super().__init__()
        self.dim = dim
        self.window_size = to_2tuple(window_size)
        win_h, win_w = self.window_size
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5
        # define a parameter table of relative position bias
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * win_h - 1) * (2 * win_w - 1), num_heads)
        )   # [2*Wh-1 * 2*Ww-1, nHeads]   Offset Range: -Wh+1, Wh-1

        self.register_buffer("relative_position_index",
                             self.get_relative_position_index(win_h, win_w), persistent=False)
        trunc_normal_(self.relative_position_bias_table, std=.02)
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attention_dropout_ratio)
        self.proj = nn.Linear(dim, dim, bias=proj_bias)
        self.proj_drop = nn.Dropout(proj_drop)

        self.softmax = nn.Softmax(dim=-1)

    def get_relative_position_index(self, win_h: int, win_w: int):
        # get pair-wise relative position index for each token inside the window
        coords = torch.stack(torch.meshgrid(torch.arange(win_h), torch.arange(win_w), indexing='ij'))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += win_h - 1  # shift to start from 0
        relative_coords[:, :, 1] += win_w - 1
        relative_coords[:, :, 0] *= 2 * win_w - 1
        return relative_coords.sum(-1)  # Wh*Ww, Wh*Ww

    def forward(self, x, mask=None):
        """
        Args:
            x: input features with shape of (num_windows*B, N, C)
            mask: (0/-inf) mask with shape of (num_windows, Wh*Ww, Wh*Ww) or None
        """
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[:3]

        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))

        relative_position_bias = self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
            self.window_size[0] * self.window_size[1], self.window_size[0] * self.window_size[1], -1)  # Wh*Ww,Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
        attn = attn + relative_position_bias.unsqueeze(0)

        if mask is not None:
            nW = mask.shape[0]
            attn = attn.view(B // nW, nW, self.num_heads, N, N) + mask.unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N, N)
            attn = self.softmax(attn)
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


def window_partition(x, window_size: int):
    """
    将feature map按照window_size划分成一个个没有重叠的window
    Args:
        x: (B, H, W, C)
        window_size (int): window size(M)

    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """
    B, H, W, C = x.shape
    x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
    # permute: [B, H//Mh, Mh, W//Mw, Mw, C] -> [B, H//Mh, W//Mh, Mw, Mw, C]
    # view: [B, H//Mh, W//Mw, Mh, Mw, C] -> [B*num_windows, Mh, Mw, C]
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
    return windows


def window_reverse(windows, window_size: int, H: int, W: int):
    """
    将一个个window还原成一个feature map
    Args:
        windows: (num_windows*B, window_size, window_size, C)
        window_size (int): Window size(M)
        H (int): Height of image
        W (int): Width of image

    Returns:
        x: (B, H, W, C)
    """
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    # view: [B*num_windows, Mh, Mw, C] -> [B, H//Mh, W//Mw, Mh, Mw, C]
    x = windows.view(B, H // window_size, W // window_size, window_size, window_size, -1)
    # permute: [B, H//Mh, W//Mw, Mh, Mw, C] -> [B, H//Mh, Mh, W//Mw, Mw, C]
    # view: [B, H//Mh, Mh, W//Mw, Mw, C] -> [B, H, W, C]
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x

class SwinTransformerBlock(nn.Module):
    r""" Swin Transformer Block."""
    mlp_ratio = 4
    def __init__(
            self,
            dim,
            num_heads,
            window_size=7,
            shift_size=0,
            qkv_bias=True,
            proj_bias=True,
            attention_dropout_ratio=0.,
            proj_drop=0.,
            drop_path_ratio=0.,
            norm_layer=LayerNorm,
            act_layer=nn.GELU,
    ):
        super(SwinTransformerBlock, self).__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        assert 0 <= self.shift_size < window_size, "shift_size must in 0-window_size"

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention(
            dim,
            window_size=self.window_size,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            proj_bias=proj_bias,
            attention_dropout_ratio=attention_dropout_ratio,
            proj_drop=proj_drop,
        )

        self.drop_path = DropPath(drop_path_ratio) if drop_path_ratio > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * self.mlp_ratio)
        self.mlp = MLP(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop_ratio=proj_bias)

        self.H = None
        self.W = None

    def forward(self, x, mask_matrix):
        """
        Args:
            x: Input feature, tensor size (B, H*W, C).
            H, W: Spatial resolution of the input feature.
            mask_matrix: Attention mask for cyclic shift.
        """
        B, L, C = x.shape
        H, W = self.H, self.W
        assert L == H * W, "input feature has wrong size"

        shortcut = x
        x = self.norm1(x)
        x = x.view(B, H, W, C)

        # pad feature maps to multiples of window size
        pad_l = pad_t = 0
        pad_r = (self.window_size - W % self.window_size) % self.window_size
        pad_b = (self.window_size - H % self.window_size) % self.window_size
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        _, Hp, Wp, _ = x.shape

        # cyclic shift
        if self.shift_size > 0:
            shifted_x = torch.roll(x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2))
            attn_mask = mask_matrix
        else:
            shifted_x = x
            attn_mask = None

        # partition windows
        x_windows = window_partition(shifted_x, self.window_size)  # nW*B, window_size, window_size, C
        x_windows = x_windows.view(-1, self.window_size * self.window_size, C)  # nW*B, window_size*window_size, C

        # W-MSA/SW-MSA
        attn_windows = self.attn(x_windows, mask=attn_mask)  # nW*B, window_size*window_size, C

        # merge windows
        attn_windows = attn_windows.view(-1, self.window_size, self.window_size, C)
        shifted_x = window_reverse(attn_windows, self.window_size, Hp, Wp)  # B H' W' C

        # reverse cyclic shift
        if self.shift_size > 0:
            x = torch.roll(shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2))
        else:
            x = shifted_x

        if pad_r > 0 or pad_b > 0:
            x = x[:, :H, :W, :].contiguous()

        x = x.view(B, H * W, C)

        # FFN
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class BasicLayer(nn.Module):
    """ A basic Swin Transformer layer for one stage."""
    def __init__(self,
                 dim,
                 num_layers,
                 num_heads,
                 drop_path,
                 window_size=7,
                 qkv_bias=True,
                 proj_bias=True,
                 attention_dropout_ratio=0.,
                 proj_drop=0.,
                 norm_layer=LayerNorm,
                 act_layer=nn.GELU,
                 downsample=None):
        super().__init__()
        self.window_size = window_size
        self.shift_size = window_size // 2
        self.num_layers = num_layers

        # build blocks
        self.blocks = nn.ModuleList([
            SwinTransformerBlock(
                dim=dim,
                num_heads=num_heads,
                window_size=window_size,
                shift_size=0 if (i % 2 == 0) else window_size // 2,
                qkv_bias=qkv_bias,
                proj_bias=proj_bias,
                attention_dropout_ratio=attention_dropout_ratio,
                proj_drop=proj_drop,
                drop_path_ratio=drop_path[i] if isinstance(drop_path, list) else drop_path,
                norm_layer=norm_layer,
                act_layer=act_layer)
            for i in range(num_layers)])
        # patch merging layer
        if downsample is not None:
            self.downsample = downsample(dim=dim, norm_layer=norm_layer)
        else:
            self.downsample = None

    def forward(self, x, H, W):
        """
        Args:
            x: Input feature, tensor size (B, H*W, C).
            H, W: Spatial resolution of the input feature.
        """

        # calculate attention mask for SW-MSA
        Hp = int(np.ceil(H / self.window_size)) * self.window_size
        Wp = int(np.ceil(W / self.window_size)) * self.window_size
        img_mask = torch.zeros((1, Hp, Wp, 1), device=x.device)  # 1 Hp Wp 1
        h_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        w_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        cnt = 0
        for h in h_slices:
            for w in w_slices:
                img_mask[:, h, w, :] = cnt
                cnt += 1

        mask_windows = window_partition(img_mask, self.window_size)  # nW, window_size, window_size, 1
        mask_windows = mask_windows.view(-1, self.window_size * self.window_size)
        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))

        for blk in self.blocks:
            blk.H, blk.W = H, W
            x = blk(x, attn_mask)
        if self.downsample is not None:
            x = self.downsample(x, H, W)
            H, W = (H + 1) // 2, (W + 1) // 2

        return x, H, W




class SwinTransformer(nn.Module):
    """ Swin Transformer backbone."""
    def __init__(self,
                 patch_size=4,
                 in_channels=3,
                 num_classes=1000,
                 embed_dim=96,
                 depths=(2, 2, 6, 2),
                 num_heads=(3, 6, 12, 24),
                 window_size=7,
                 qkv_bias=True,
                 proj_bias=True,
                 attention_dropout_ratio=0.,
                 proj_drop=0.,
                 drop_path_rate=0.2,
                 norm_layer=LayerNorm,
                 patch_norm=True,
                 ):
        super().__init__()
        self.num_classes = num_classes
        self.num_layers = len(depths)
        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.patch_norm = patch_norm
        # stage4输出特征矩阵的channels
        self.num_features = int(embed_dim * 2 ** (self.num_layers - 1))

        # split image into non-overlapping patches
        self.patch_embed = PatchPartition(
            patch_size=patch_size, in_channels=in_channels, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)

        self.pos_drop = nn.Dropout(p=proj_drop)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        layers = []
        for i_layer in range(self.num_layers):
            layer = BasicLayer(
                dim=int(embed_dim * 2 ** i_layer),
                num_layers=depths[i_layer],
                num_heads=num_heads[i_layer],
                window_size=window_size,
                qkv_bias=qkv_bias,
                proj_bias=proj_bias,
                attention_dropout_ratio=attention_dropout_ratio,
                proj_drop=proj_drop,
                drop_path=dpr[sum(depths[:i_layer]):sum(depths[:i_layer + 1])],
                norm_layer=norm_layer,
                downsample=PatchMerging if (i_layer < self.num_layers - 1) else None,
            )
            layers.append(layer)

        self.layers = nn.Sequential(*layers)

        self.norm = norm_layer(self.num_features)
        self.avgpool = nn.AdaptiveAvgPool1d(1)
        self.head = nn.Linear(self.num_features, num_classes) if num_classes > 0 else nn.Identity()

        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

    def forward(self, x):
        # x: [B, L, C]
        x, H, W = self.patch_embed(x)
        x = self.pos_drop(x)

        for layer in self.layers:
            x, H, W = layer(x, H, W)

        x = self.norm(x)  # [B, L, C]
        x = self.avgpool(x.transpose(1, 2))
        x = torch.flatten(x, 1)
        x = self.head(x)
        return x

def swin_t(num_classes) -> SwinTransformer:
    model = SwinTransformer(in_channels=3,
                            patch_size=4,
                            window_size=7,
                            embed_dim=96,
                            depths=(2, 2, 6, 2),
                            num_heads=(3, 6, 12, 24),
                            num_classes=num_classes)
    return model

def swin_s(num_classes) -> SwinTransformer:
    model = SwinTransformer(in_channels=3,
                            patch_size=4,
                            window_size=7,
                            embed_dim=96,
                            depths=(2, 2, 18, 2),
                            num_heads=(3, 6, 12, 24),
                            num_classes=num_classes)
    return model


def swin_b(num_classes) -> SwinTransformer:
    model = SwinTransformer(in_channels=3,
                            patch_size=4,
                            window_size=7,
                            embed_dim=128,
                            depths=(2, 2, 18, 2),
                            num_heads=(4, 8, 16, 32),
                            num_classes=num_classes)
    return model

def swin_l(num_classes) -> SwinTransformer:
    model = SwinTransformer(in_channels=3,
                            patch_size=4,
                            window_size=7,
                            embed_dim=192,
                            depths=(2, 2, 18, 2),
                            num_heads=(6, 12, 24, 48),
                            num_classes=num_classes)
    return model

if __name__=="__main__":
    import pyzjr
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = swin_l(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    pyzjr.summary_1(net, input_size=(3, 224, 224))
    # swin_t Total params: 27,499,108
    # swin_s Total params: 48,792,676
    # swin_b Total params: 86,683,780
    # swin_l Total params: 194,906,308


from .loss_function import (
    L1Loss,
    L2Loss,
    BCELoss,
    Joint2loss,
    MCCLoss
)

from .dice_focal import (
    DiceLoss,
    FocalLoss,
    DiceFocalLoss,
    JaccardLoss
)

from .constants import (
    BINARY_MODE,
    MULTILABEL_MODE,
    MULTICLASS_MODE
)


import torch.nn as nn

class ConvNormAct(nn.Module):
    def __init__(
            self,
            in_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            dilation=1,
            groups=1,
            bias=False,
            apply_act=True,
            norm_layer=nn.BatchNorm2d,
            act_layer=nn.ReLU(inplace=True),
            **kwargs,
    ):
        super(ConvNormAct, self).__init__()
        if padding is None:
            padding = (kernel_size - 1) / 2
        self.apply_act = apply_act
        self.conv = nn.Conv2d(
            in_channels, out_channels, kernel_size, stride=stride,
            padding=padding, dilation=dilation, groups=groups, bias=bias, **kwargs)
        self.bn = norm_layer(out_channels)
        self.act = act_layer

    @property
    def in_channels(self):
        return self.conv.in_channels

    @property
    def out_channels(self):
        return self.conv.out_channels

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        if self.apply_act:
            x = self.act(x)
        return x


class ConvNorm(ConvNormAct):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size=3,
                 stride=1,
                 padding=1,
                 dilation=1,
                 groups=1,
                 bias=False,
                 norm_layer=nn.BatchNorm2d,
                 **kwargs,):
        super().__init__(in_channels=in_channels,
                         out_channels=out_channels,
                         kernel_size=kernel_size,
                         stride=stride,
                         padding=padding,
                         dilation=dilation,
                         groups=groups,
                         bias=bias,
                         apply_act=False,
                         norm_layer=norm_layer,
                         act_layer=nn.ReLU(inplace=True),  # ignore, actually not working
                         **kwargs,)

class NormAct(nn.Module):
    def __init__(self,
                 out_channels,
                 norm_layer=nn.BatchNorm2d,
                 act_layer=nn.ReLU(inplace=True),
                 ):
        super(NormAct, self).__init__()
        self.norm_layer = norm_layer(out_channels)
        self.act_layer = act_layer

    def forward(self, x):
        x = self.norm_layer(x)
        x = self.act_layer(x)
        return x

class ConvAct(nn.Module):
    def __init__(
            self,
            in_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            dilation=1,
            groups=1,
            bias=False,
            act_layer=nn.ReLU(inplace=True),
    ):
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride,
                              padding, dilation, groups, bias)
        self.activation = act_layer

    @property
    def in_channels(self):
        return self.conv.in_channels

    @property
    def out_channels(self):
        return self.conv.out_channels

    def forward(self, x):
        x = self.conv(x)
        x = self.activation(x)
        return x

def conv3x3(in_channels, out_channels, stride=1, groups=1, padding=1, dilation=1, bias=False):
    """3x3 convolution"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride,
                     padding=padding, groups=groups, bias=bias, dilation=dilation)

def conv1x1(in_channels, out_channels, stride=1, bias=False):
    """1x1 convolution"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=bias)

ConvBnReLU = ConvBNReLU = ConvNormAct
ConvBn = ConvBN = ConvNorm
BnReLU = BNReLU = NormAct
ConvReLU = ConvAct


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is used for visualizing useful functions.
"""
import cv2
import time
import numpy as np
from functools import wraps

class FPS:
    """
    Detect video frame rate and refresh the video display
    Examples:
    ```
        fpsReader = FPS()
        Vcap = VideoCap(mode=0)
        while True:
            img = Vcap.read()
            fps, img = fpsReader.update(img)
            Vcap.show("ss", img)
    ```
    """
    def __init__(self):
        self.pTime = time.time()

    def update(self, img=None, pos=(20, 50), color=(255, 0, 0), scale=3, thickness=3):
        """
        Update frame rate
        :param img: The displayed image can be left blank if only the fps value is needed
        :param pos: Position on FPS on image
        :param color: The color of the displayed FPS value
        :param scale: The proportion of displayed FPS values
        :param thickness: The thickness of the displayed FPS value
        """
        cTime = time.time()
        try:
            fps = 1 / (cTime - self.pTime)
            self.pTime = cTime
            if img is None:
                return fps
            else:
                cv2.putText(img, f'FPS: {int(fps)}', pos, cv2.FONT_HERSHEY_PLAIN,
                            scale, color, thickness)
                return fps, img
        except:
            return 0

class Timer:
    def __init__(self):
        """Start is called upon creation"""
        self.times = []
        self.start()

    def start(self):
        """initial time"""
        self.gap = time.time()

    def stop(self):
        """The time interval from the start of timing to calling the stop method"""
        self.times.append(time.time() - self.gap)
        return self.times[-1]

    def avg(self):
        """Average time consumption"""
        return sum(self.times) / len(self.times)

    def total(self):
        """Total time consumption"""
        return sum(self.times)

    def cumsum(self):
        """Accumulated sum of time from the previous n runs"""
        return np.array(self.times).cumsum().tolist()


class Runcodes:
    """
    Comparing the running time of different algorithms.
    Examples:
    ```
        with Runcodes("inference time"):
            output = ...
    ```
    """
    def __init__(self, description='Done'):
        self.description = description

    def __enter__(self):
        self.timer = Timer()
        return self

    def __exit__(self, *args):
        print(f'{self.description}: {self.timer.stop():.7f} sec')


def timing(decimal=5):
    """计时器装饰器，用于测量函数执行的时间。"""
    def decorator(function):
        @wraps(function)
        def timingwrap(*args, **kwargs):
            print(function.__name__, flush=True)
            start = time.perf_counter()
            res = function(*args, **kwargs)
            end = time.perf_counter()
            execution_time = end - start
            format_string = "{:.{}f}".format(execution_time, decimal)
            print(function.__name__, "delta time (s) =", format_string, flush=True)
            return res
        return timingwrap
    return decorator

if __name__=="__main__":
    @timing(decimal=5)
    def test_function():
        time.sleep(2.5)
    test_function()


"""
Copyright (c) 2025, Auorui.
All rights reserved.

This module is used for reading and displaying images and videos. The image contains
OpenCV and PIL, and the loading method for the video is OpenCV.
"""
from .imvision import bgr_read, imwriter, display, url2image, stacked2image, \
    StackedImagesV1, StackedImagesV2, imattributes
from .pilvision import rgb_read, imshowplt, StackedpltV1, StackedpltV2, matplotlib_patch
from .videovision import VideoCap, Mp4toGif, FindColor, DetectImageColor, DetectVideoColor


import os
from pyzjr.data.base import natsorted
from pyzjr.utils.check import is_str, is_list_or_tuple

def generate_txt(target_path, onlybasename=False, txt_path=r'./output.txt'):
    """将指定文件夹下的文件路径（或文件名）写入到指定的文本文件中"""
    f = open(txt_path, "w")
    files = natsorted(os.listdir(target_path))
    files_num = len(files)
    index_count = 0
    count = 0
    for file in files:
        index_count = index_count + 1
        path = os.path.splitext(file)[0] if onlybasename else os.path.join(target_path, str(file))
        if count == files_num - 1:
            f.write(path)
            break
        if index_count >= 0:
            f.write(path + "\n")
            count = count + 1
    f.close()

def read_txt(txt_path):
    """从读取的文件中获取列表"""
    path_list = []
    if isinstance(txt_path, str):  # Check if txt_path is a string
        with open(txt_path, 'r') as file:
            lines = file.readlines()
            path_list.extend(line.strip() for line in lines)
    return path_list

def write_to_txt(file_path, content, encoding='utf-8'):
    """将指定的内容写入到指定的文本文件中"""
    if os.path.exists(file_path):
        os.remove(file_path)

    if is_list_or_tuple(content):
        with open(file_path, 'a', encoding=encoding) as file:
            for i in content:
                if is_str(i):
                    file.write(i)
                else:
                    file.write(str(i))
                file.write("\n")

    elif is_str(content):
        with open(file_path, 'w', encoding=encoding) as file:
            file.write(content)

if __name__ == "__main__":
    file_dir = r"D:\PythonProject\pyzjrPyPi\pyzjr\data"
    generate_txt(file_dir, onlybasename=True)
    print(read_txt("./output.txt"))


"""
Copyright (c) 2023, Auorui.
All rights reserved.
"""
import cv2
import numpy as np

def OverlayPng(imgBack, imgFront, pos=(0, 0), alpha_gain=1.0):
    """
    Overlay display image with proper alpha blending
    :param imgBack: Background image, no format requirement, 3 channels
    :param imgFront: PNG pre image, must be read using cv2.IMREAD_UNCHANGED=-1
    :param pos: Placement position

    Examples:
    '''
        background = cv2.imread(background_path)
        overlay = cv2.imread(overlay_path, cv2.IMREAD_UNCHANGED)
        fused_image = pyzjr.OverlayPng(background, overlay, alpha_gain=1.5)
    '''
    """
    hf, wf, cf = imgFront.shape
    hb, wb, cb = imgBack.shape
    y_pos, x_pos = pos
    y_end = y_pos + hf
    x_end = x_pos + wf

    # Ensure we don't go beyond the background boundaries
    if y_end > hb:
        y_end = hb
    if x_end > wb:
        x_end = wb

    # Resize overlay to fit the background (optional but good practice)
    overlay_resized = cv2.resize(imgFront, (x_end - x_pos, y_end - y_pos))

    overlay_alpha = overlay_resized[:, :, 3].astype(float) / 255.0
    overlay_alpha = np.clip(overlay_alpha * alpha_gain, 0, 1)
    background_alpha = 1.0 - overlay_alpha

    result = overlay_resized[:, :, :3] * overlay_alpha[..., np.newaxis] + imgBack[y_pos:y_end, x_pos:x_end, :3] * background_alpha[..., np.newaxis]
    imgBack[y_pos:y_end, x_pos:x_end, :3] = result.astype(np.uint8)

    return imgBack


def ConvertBbox(bbox, to_center=True):
    """
    Convert bounding box coordinates between (upper left, lower right) and (center, width, height).

    :param bbox: List of bounding boxes in either [x1, y1, x2, y2] or [cx, cy, w, h] format.
    :param to_center: If True, convert from (upper left, lower right) to (center, width, height).
                      If False, convert from (center, width, height) to (upper left, lower right).
    :return: List of converted bounding boxes.
    """
    converted_boxes = []
    for box in bbox:
        if to_center:
            # Convert from (upper left, lower right) to (center, width, height)
            x1, y1, x2, y2 = box
            cx = (x1 + x2) / 2
            cy = (y1 + y2) / 2
            w = x2 - x1
            h = y2 - y1
            converted_box = [cx, cy, w, h]
        else:
            # Convert from (center, width, height) to (upper left, lower right)
            cx, cy, w, h = box
            x1 = cx - 0.5 * w
            y1 = cy - 0.5 * h
            x2 = cx + 0.5 * w
            y2 = cy + 0.5 * h
            converted_box = [x1, y1, x2, y2]
        converted_boxes.append(converted_box)
    return converted_boxes



"""
数据集标准如下:
    - base                          - 数据集名称
        - train                         - 训练集
            - crack                         - 类别名 1
            - rust                          - 类别名 2
            - spalling                      - 类别名 3
            - stoma                         - 类别名 4
        - val                           - 验证集
        - test                          - 测试集（ 不一定需要 ）
"""
import os
import cv2
import torch
import random
import numpy as np
from pathlib import Path
from natsort import natsorted
from pyzjr.data.datasets.baseset import BaseDataset

class ClassificationDataset(BaseDataset):
    def __init__(self, root_dir, target_shape, is_train=True, transform=None):
        super(ClassificationDataset, self).__init__()
        data_folder = os.path.join(root_dir, 'train') if is_train else os.path.join(root_dir, 'val')
        self.categories = natsorted(os.listdir(data_folder))
        self.num_classes = len(self.categories)
        if isinstance(target_shape, (list, tuple)) and len(target_shape) == 2:
            self.target_shape = target_shape
        elif isinstance(target_shape, int):
            self.target_shape = (target_shape, target_shape)
        self.transform = transform
        self.data_and_label_list = []
        self.is_train = is_train
        for label, category in enumerate(self.categories):
            category_folder = os.path.join(data_folder, category)
            self.data_and_label_list.extend(
                [(str(f), label) for f in Path(category_folder).iterdir() if f.is_file() and not f.name.startswith(".")])

    def __len__(self):
        return len(self.data_and_label_list)

    def __getitem__(self, index):
        cv2.setNumThreads(0)
        cv2.ocl.setUseOpenCL(False)
        image_path, label = self.data_and_label_list[index]
        label = torch.tensor(label).long()
        image = cv2.imread(image_path, cv2.IMREAD_COLOR)
        if self.transform:
            image = self.transform(image)
        else:
            image = self.class_augument(image, self.target_shape, is_train=self.is_train)
            image = np.transpose(image, [2, 0, 1])
            image = torch.from_numpy(image).float()
        return image, label

    def class_augument(self, image, target_shape, is_train, prob=.5):
        h, w = target_shape
        ih, iw = image.shape[:2]
        scale = min(w / iw, h / ih)
        nw = int(iw * scale)
        nh = int(ih * scale)
        resized_image = cv2.resize(image, (nw, nh), interpolation=cv2.INTER_CUBIC)
        new_image = np.full((h, w, 3), (128, 128, 128), dtype=np.uint8)
        top = (h - nh) // 2
        left = (w - nw) // 2
        new_image[top:top + nh, left:left + nw] = resized_image
        new_image = cv2.cvtColor(new_image, cv2.COLOR_BGR2RGB)
        if is_train:
            if random.random() > (1 - prob):
                new_image = np.flip(new_image, axis=1)
            r = random.randint(0, 3)
            new_image = np.rot90(new_image, r, (0, 1))
            # 转换到HSV颜色空间
            hsv_image = cv2.cvtColor(new_image, cv2.COLOR_RGB2HSV).astype(np.float32)
            h_channel, s_channel, v_channel = cv2.split(hsv_image)
            # 随机调整色调（Hue）
            delta_h = random.randint(-30, 30)
            h_channel = (h_channel + delta_h) % 180
            # 随机调整饱和度（Saturation）和亮度（Value）
            s_scale = np.random.uniform(0.5, 1.5)
            v_scale = random.uniform(0.5, 1.5)
            s_channel = np.clip(s_channel * s_scale, 0, 255).astype(np.uint8)
            v_channel = np.clip(v_channel * v_scale, 0, 255).astype(np.uint8)
            # 合并通道并转换回RGB颜色空间
            hsv_image[..., 0] = h_channel.astype(np.uint8)
            hsv_image[..., 1] = s_channel
            hsv_image[..., 2] = v_channel
            new_image = cv2.cvtColor(hsv_image.astype(np.uint8), cv2.COLOR_HSV2RGB)
        # 将图像归一化到[0, 1]范围
        new_image = new_image / 255.0
        return new_image


if __name__=="__main__":
    from torchvision import transforms
    from torch.utils.data import DataLoader
    data_path = r'E:\PythonProject\pytorch_classification_Auorui\data\cat_dog'
    train_dataset = ClassificationDataset(root_dir=data_path, target_shape=[256, 256], is_train=True)
    val_dataset = ClassificationDataset(root_dir=data_path, target_shape=[256, 256], is_train=False)
    print("训练集数量：", len(train_dataset), "训练集数量", len(val_dataset))

    train_loader = DataLoader(train_dataset, shuffle=True, batch_size=4, num_workers=2, pin_memory=True)
    val_loader = DataLoader(val_dataset, batch_size=1, num_workers=2, pin_memory=True)
    # for batch in val_dataset:
    #     image, label = batch
    #     print(image.shape, label)
    # show_image_from_dataloader(train_dataset)


    transform = transforms.Compose([
        transforms.ToPILImage(),  # convert numpy to PIL image
        transforms.RandomHorizontalFlip(p=0.5),
        transforms.RandomVerticalFlip(p=0.5),
        transforms.RandomRotation(degrees=90),
        transforms.ColorJitter(brightness=0.2, contrast=0.2, saturation=0.2, hue=0.1),
        transforms.RandomResizedCrop(size=(256, 256)),
        transforms.PILToTensor()
    ])
    train_dataset = ClassificationDataset(root_dir=data_path, target_shape=[256, 256], is_train=True,
                                          transform=transform)
    # show_image_from_dataloader(train_dataset)
    # for batch in train_dataset:
    #     image, label = batch
    #     print(image.shape, label)

    def show_image_from_dataloader(test_dataset):
        loader = DataLoader(
            test_dataset,
            batch_size=1,
            shuffle=True,
            num_workers=1,
            pin_memory=False,
            drop_last=True
        )
        for i, (img, label) in enumerate(loader):
            img = img[0].numpy()
            img = np.transpose(img, (1, 2, 0))
            img = (img*255).astype(np.uint8)
            img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
            cv2.putText(img, f"{label.item()}", (30, 30), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale=1, thickness=2, color=(255, 0, 255))
            cv2.imshow('show image from dataloader', img)
            cv2.waitKey(0)
            cv2.destroyAllWindows()


# import pyzjr.Z as Z
# -i https://pypi.tuna.tsinghua.edu.cn/simple
import numpy as np

def getColor(color_name, to_rgb=False):
    bgr_colors = {
        'blue': (255, 0, 0),
        'green': (0, 255, 0),
        'red': (0, 0, 255),
        "dark_blue": (128, 0, 0),
        'dark_green': (0, 128, 0),
        "dark_red": (0, 0, 128),
        "blue_green": (128, 128, 0),
        'magenta': (255, 0, 255),
        "black": (0, 0, 0),
        "grey": (128, 128, 128),
        "silvery": (192, 192, 192),
        "white": (255, 255, 255),
        "yellow": (0, 255, 255),
        "orange": (0, 97, 255),
        "purple": (255, 0, 255),
        "violet": (240, 32, 160),
        "brown": (19, 69, 139),
        "pink": (203, 192, 255),
    }
    bgr_color = bgr_colors.get(color_name, None)
    if bgr_color is not None and to_rgb:
        return bgr_color[::-1]
    return bgr_color

e = 2.718281828459045
pi = 3.141592653589793
half_pi = pi / 2
double_pi = 2 * pi

NUMPY_DTYPE = {
    np.dtype("uint8"): 255,
    np.dtype("uint16"): 65535,
    np.dtype("uint32"): 4294967295,
    np.dtype("float32"): 1.0,
}

IMG_FORMATS = ('bmp', 'dng', 'jpeg', 'jpg', 'mpo', 'png', 'tif', 'tiff', 'webp', 'pfm')  # include image suffixes
VID_FORMATS = ('asf', 'avi', 'gif', 'm4v', 'mkv', 'mov', 'mp4', 'mpeg', 'mpg', 'ts', 'wmv')  # include video suffixes

# 颜色空间转换
BGR2RGB = 4
BGR2HSV = 40
BGR2GRAY = 6
RGB2GRAY = 7
GRAY2BGR = 8
GRAY2RGB = 8
HSV2BGR = 54
HSV2RGB = 55
RGB2HSV = 41
RGB2BGR = 4

# video
Esc = 27

# BGR
blue = getColor('blue')
green = getColor('green')
red = getColor('red')
dark_blue = getColor('dark_blue')
dark_green = getColor('dark_green')
dark_red = getColor('dark_red')
blue_green = getColor('blue_green')
magenta = getColor('magenta')
black = getColor('black')
grey = getColor('grey')
silvery = getColor('silvery')
white = getColor('white')
yellow = getColor('yellow')
orange = getColor('orange')
purple = getColor('purple')
violet = getColor('violet')
brown = getColor('brown')
pink = getColor('pink')
# RGB
rgb_blue = getColor('blue', to_rgb=True)
rgb_green = getColor('green', to_rgb=True)
rgb_red = getColor('red', to_rgb=True)
rgb_dark_blue = getColor('dark_blue', to_rgb=True)
rgb_dark_green = getColor('dark_green', to_rgb=True)
rgb_dark_red = getColor('dark_red', to_rgb=True)
rgb_blue_green = getColor('blue_green', to_rgb=True)
rgb_magenta = getColor('magenta', to_rgb=True)
rgb_black = getColor('black', to_rgb=True)
rgb_grey = getColor('grey', to_rgb=True)
rgb_silvery = getColor('silvery', to_rgb=True)
rgb_white = getColor('white', to_rgb=True)
rgb_yellow = getColor('yellow', to_rgb=True)
rgb_orange = getColor('orange', to_rgb=True)
rgb_purple = getColor('purple', to_rgb=True)
rgb_violet = getColor('violet', to_rgb=True)
rgb_brown = getColor('brown', to_rgb=True)
rgb_pink = getColor('pink', to_rgb=True)


if __name__=="__main__":
    import cv2
    import numpy as np

    height, width = 300, 400
    image = np.zeros((height, width, 3), dtype=np.uint8)
    image[:] = (255, 0, 255)
    cv2.imshow('Color', image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


import torch
import torch.nn as nn
import torch.nn.functional as F
from pyzjr.nn.losses.constants import BINARY_MODE, MULTICLASS_MODE, MULTILABEL_MODE

class DiceLoss(nn.Module):
    """
    Dice Loss 测量预测的和目标的二进制分割掩码之间的不相似性。它被计算为1减去Dice系数，Dice系数是重叠的度量
    在预测区域和目标区域之间。

    Args:
        mode (str): Loss 模式，可选 'binary', 'multiclass' 或 'multilabel'。
        reduction (str, optional): 指定对输出应用的缩减方式。
            可选值为 'none', 'mean', 或 'sum'。默认为 'mean'。

    Returns:
        torch.Tensor: The Dice Loss between input and target.
    """
    def __init__(self, mode=MULTILABEL_MODE, ignore_index=None, reduction='mean', eps=1e-5):
        super(DiceLoss, self).__init__()
        self.mode = mode
        self.eps = eps
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:,:self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]

        if self.mode == BINARY_MODE:
            input = torch.sigmoid(input)
            intersection = torch.sum(input * target)
            union = torch.sum(input) + torch.sum(target)
        elif self.mode == MULTICLASS_MODE:
            input = torch.softmax(input, dim=1)
            target_one_hot = F.one_hot(target, num_classes=input.shape[1]).permute(0, 3, 1, 2).float()
            intersection = torch.sum(input * target_one_hot, dim=(2, 3))
            union = torch.sum(input, dim=(2, 3)) + torch.sum(target_one_hot, dim=(2, 3))
        elif self.mode == MULTILABEL_MODE:
            intersection = torch.sum(input * target, dim=(1,))
            union = torch.sum(input, dim=(1,)) + torch.sum(target, dim=(1,))

        dice_loss = 1 - (2 * intersection + self.eps) / (union + self.eps)
        return self.aggregate_loss(dice_loss)

    def aggregate_loss(self, focal_loss: torch.Tensor) -> torch.Tensor:
        if self.reduction == 'mean':
            return focal_loss.mean()
        elif self.reduction == 'sum':
            return focal_loss.sum()
        elif self.reduction == 'none':
            return focal_loss
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')

class FocalLoss(nn.Module):
    """
    Focal Loss 用于解决类别不平衡问题，通过缩小易分类的类别的损失来关注难分类的类别。依据公式实现。

    Args:
        mode (str): Loss 模式，可选 'binary', 'multiclass' 或 'multilabel'。
        alpha (float, optional): 控制易分类的类别的权重，大于1表示增加权重，小于1表示减小权重。默认为1.
        gamma (float, optional): 控制难分类的类别的损失的下降速度，大于0表示下降较慢，小于0表示下降较快。默认为2.
        reduction (str, optional): 指定对输出应用的缩减方式。
            可选值为 'none', 'mean', 或 'sum'。默认为 'mean'。
    """
    def __init__(self,
                 mode: str = MULTILABEL_MODE,
                 alpha: float = 1,
                 gamma: float = 2,
                 ignore_index: int = None,
                 reduction: str = 'mean',
                 ):
        super(FocalLoss, self).__init__()
        assert mode in {BINARY_MODE, MULTILABEL_MODE, MULTICLASS_MODE}
        self.mode = mode
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if self.ignore_index:
            input = input[:,:self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]
        if self.mode == BINARY_MODE:
            input = torch.sigmoid(input)
            ce_loss = F.binary_cross_entropy_with_logits(input, target.float(), reduction='none')
            class_weights = torch.exp(-ce_loss)
            focal_loss = self.alpha * (1 - class_weights) ** self.gamma * ce_loss
        elif self.mode == MULTICLASS_MODE:
            ce_loss = F.cross_entropy(input, target.long(), reduction='none')
            class_weights = torch.exp(-ce_loss)
            focal_loss = self.alpha * (1 - class_weights) ** self.gamma * ce_loss
        elif self.mode == MULTILABEL_MODE:
            ce_loss = F.binary_cross_entropy_with_logits(input, target.float(), reduction='none')
            class_weights = torch.exp(-ce_loss)
            focal_loss = self.alpha * (1 - class_weights) ** self.gamma * ce_loss

        return self.aggregate_loss(focal_loss)

    def aggregate_loss(self, focal_loss: torch.Tensor) -> torch.Tensor:
        if self.reduction == 'mean':
            return focal_loss.mean()
        elif self.reduction == 'sum':
            return focal_loss.sum()
        elif self.reduction == 'none':
            return focal_loss
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')

class DiceFocalLoss(nn.Module):
    """
    DiceFocalLoss结合了Dice Loss和Focal Loss两种损失函数，用于解决图像分割任务中的类别不平衡和边界模糊的问题。
    在计算损失时，综合考虑了模型预测结果与真实标签的 Dice Loss 和 Focal Loss 。

    Args:
        ignore_index (int, optional): 需要忽略的标签索引，如果设置了该参数，则计算损失时会忽略这些标签。默认为None。
        reduction (str, optional): 损失值的缩减模式，可选值为'mean'（求平均）、'sum'（求和）或'none'（不缩减）。默认为'mean'。
        eps (float, optional): 用于数值稳定性的小值。默认为1e-5。
        lambda_dice (float, optional): Dice Loss的权重系数。默认为1.0。
        lambda_focal (float, optional): Focal Loss的权重系数。默认为1.0。

    Examples::
        >>> criterion = DiceFocalLoss(ignore_index=0, reduction='mean', lambda_dice=0.8, lambda_focal=0.2)
        >>> input_data = torch.rand((4, 2, 16, 16), dtype=torch.float32)
        >>> target_data = torch.randint(0, 2, (4, 2, 16, 16), dtype=torch.float32)
        >>> loss = criterion(input_data, target_data)
        >>> print("DiceFocal Loss:", loss.item())

    Returns:
        torch.Tensor: 计算得到的DiceFocal Loss值。
    """
    def __init__(self,
                 mode=MULTILABEL_MODE ,
                 ignore_index=None,
                 reduction='mean',
                 eps=1e-5,
                 lambda_dice=1.0,
                 lambda_focal=1.0):
        super(DiceFocalLoss, self).__init__()
        self.eps = eps
        self.reduction = reduction
        self.dice = DiceLoss(
            mode,
            ignore_index=ignore_index,
            reduction='none',
        )
        self.focal = FocalLoss(
            mode,
            ignore_index=ignore_index,
            reduction='none',
        )
        if lambda_dice < 0.0:
            raise ValueError("lambda_dice should be no less than 0.0.")
        if lambda_focal < 0.0:
            raise ValueError("lambda_focal should be no less than 0.0.")
        self.lambda_dice = lambda_dice
        self.lambda_focal = lambda_focal

    def forward(self, input, target):
        dice_loss = self.dice(input, target)
        focal_loss = self.focal(input, target)
        total_loss = self.lambda_dice * dice_loss + self.lambda_focal * focal_loss

        return self.aggregate_loss(total_loss)

    def aggregate_loss(self, total_loss: torch.Tensor) -> torch.Tensor:
        if self.reduction == 'mean':
            return total_loss.mean()
        elif self.reduction == 'sum':
            return total_loss.sum()
        elif self.reduction == 'none':
            return total_loss
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')

class JaccardLoss(nn.Module):
    """
    Jaccard Loss 用于衡量预测分割与目标分割之间的相似度。它是通过计算预测区域和目标区域的交集与并集之间的比值来衡量的。

    Args:
        mode (str): Loss 模式，可选 'binary', 'multiclass' 或 'multilabel'。
        ignore_index (int, optional): 忽略的类别索引，默认为 None。
        reduction (str, optional): 指定对输出应用的缩减方式。
            可选值为 'none', 'mean', 或 'sum'。默认为 'mean'。
        eps (float, optional): 避免除以零的小量值。默认为 1e-5。
    """
    def __init__(self, mode=MULTILABEL_MODE, ignore_index=None, reduction='mean', eps=1e-5):
        super(JaccardLoss, self).__init__()
        self.mode = mode
        self.eps = eps
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:, :self.ignore_index, ...]
            target = target[:, :self.ignore_index, ...]

        if self.mode == BINARY_MODE:
            input = torch.sigmoid(input)
            intersection = torch.sum(input * target, dim=(1,))
            union = torch.sum(input, dim=(1,)) + torch.sum(target, dim=(1,))
        elif self.mode == MULTICLASS_MODE:
            input = torch.softmax(input, dim=1)
            target_one_hot = F.one_hot(target, num_classes=input.shape[1]).permute(0, 3, 1, 2).float()
            intersection = torch.sum(input * target_one_hot, dim=(2, 3))
            union = torch.sum(input, dim=(2, 3)) + torch.sum(target_one_hot, dim=(2, 3))
        elif self.mode == MULTILABEL_MODE:
            intersection = torch.sum(input * target, dim=(1,))
            union = torch.sum(input, dim=(1,)) + torch.sum(target, dim=(1,))

        jaccard_loss = 1.0 - (intersection + self.eps) / (union - intersection + self.eps)
        return self.aggregate_loss(jaccard_loss)

    def aggregate_loss(self, jaccard_loss: torch.Tensor) -> torch.Tensor:
        if self.reduction == 'mean':
            return jaccard_loss.mean()
        elif self.reduction == 'sum':
            return jaccard_loss.sum()
        elif self.reduction == 'none':
            return jaccard_loss
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')

if __name__=="__main__":
    # Binary Mode
    criterion_binary_focal = FocalLoss(mode='binary', alpha=1, gamma=2, reduction='mean')
    criterion_binary_dice = DiceLoss(mode=BINARY_MODE, reduction='mean')
    criterion_binary_iou = JaccardLoss(mode=BINARY_MODE, reduction='mean')
    input_binary = torch.rand((2, 1, 16, 16))
    target_binary = torch.randint(0, 2, (2, 1, 16, 16))
    loss_binary_focal = criterion_binary_focal(input_binary, target_binary)
    loss_binary_dice = criterion_binary_dice(input_binary, target_binary)
    loss_binary_iou = criterion_binary_iou(input_binary, target_binary)
    print(input_binary.shape, target_binary.shape)
    print("Binary Focal Loss:", loss_binary_focal.item())
    print("Binary Dice Loss:", loss_binary_dice.item())
    print("Binary IOU Loss:", loss_binary_iou.item())

    # Multiclass Mode
    criterion_multiclass_focal = FocalLoss(mode=MULTICLASS_MODE, alpha=1, gamma=2, reduction='mean')
    criterion_multiclass_dice = DiceLoss(mode=MULTICLASS_MODE, reduction='mean')
    criterion_multiclass_iou = JaccardLoss(mode=MULTICLASS_MODE, reduction='mean')
    input_multiclass = torch.randn((2, 3, 16, 16))
    target_multiclass = torch.randint(0, 3, (2, 16, 16))
    loss_multiclass_focal = criterion_multiclass_focal(input_multiclass, target_multiclass)
    loss_multiclass_dice = criterion_multiclass_dice(input_multiclass, target_multiclass)
    loss_multiclass_iou = criterion_multiclass_iou(input_multiclass, target_multiclass)
    print(input_multiclass.shape, target_multiclass.shape)
    print("Multiclass Focal Loss:", loss_multiclass_focal.item())
    print("Multiclass Dice Loss:", loss_multiclass_dice.item())
    print("Multiclass IOU Loss:", loss_multiclass_iou.item())

    # Multilabel Mode
    criterion_multilabel_focal = FocalLoss(mode=MULTILABEL_MODE, alpha=1, gamma=2, reduction='mean')
    criterion_multilabel_dice = DiceLoss(mode=MULTILABEL_MODE, reduction='mean')
    criterion_multilabel_iou = JaccardLoss(mode=MULTILABEL_MODE, reduction='mean')
    input_multilabel = torch.randn((2, 3, 16, 16))
    target_multilabel = torch.randint(0, 2, (2, 3, 16, 16))
    loss_multilabel_focal = criterion_multilabel_focal(input_multilabel, target_multilabel)
    loss_multilabel_dice = criterion_multilabel_dice(input_multilabel, target_multilabel)
    loss_multilabel_iou = criterion_multilabel_iou(input_multilabel, target_multilabel)
    print(input_multilabel.shape, target_multilabel.shape)
    print("Multilabel Focal Loss:", loss_multilabel_focal.item())
    print("Multilabel Dice Loss:", loss_multilabel_dice.item())
    print("Multilabel IOU Loss:", loss_multilabel_iou.item())



import os
import numpy as np
from torch.utils.data import DataLoader
from pyzjr.data.datasets.baseset import BaseDataset
from pyzjr.utils.converter import to_2tuple
from pyzjr.nn.torchutils.OneHot import get_one_hot

VOC_COLOR = [[0, 0, 0], [128, 0, 0], [0, 128, 0], [128, 128, 0],
             [0, 0, 128], [128, 0, 128], [0, 128, 128], [128, 128, 128],
             [64, 0, 0], [192, 0, 0], [64, 128, 0], [192, 128, 0],
             [64, 0, 128], [192, 0, 128], [64, 128, 128], [192, 128, 128],
             [0, 64, 0], [128, 64, 0], [0, 192, 0], [128, 192, 0],
             [0, 64, 128]]

VOC_CLASSES = ['background', 'aeroplane', 'bicycle', 'bird', 'boat',
               'bottle', 'bus', 'car', 'cat', 'chair', 'cow',
               'diningtable', 'dog', 'horse', 'motorbike', 'person',
               'potted plant', 'sheep', 'sofa', 'train', 'tv/monitor']

class PascalVOCSegDataset(BaseDataset):
    def __init__(
            self,
            root="./VOCdevkit/VOC2012",
            num_classes=21,
            target_shape=256,
            is_train=True,
            class_mapping=None,
            transforms=None,
    ):
        super(PascalVOCSegDataset, self).__init__()
        txt_name = "train.txt" if is_train else "val.txt"
        self.seg_txt_path = os.path.join(root, r"ImageSets/Segmentation", txt_name)
        self.jpeg_images_path = os.path.join(root, r"JPEGImages")
        self.segmentation_class_path = os.path.join(root, r"SegmentationClass")
        self.target_shape = to_2tuple(target_shape)
        self.class_mapping = class_mapping
        self.image_names = self.read_txt(self.seg_txt_path)
        self.num_classes = num_classes
        self.transforms = transforms

    def __len__(self):
        return len(self.image_names)

    def __getitem__(self, item):
        self.disable_cv2_multithreading()
        feature_path = os.path.join(self.jpeg_images_path, f'{self.image_names[item]}.jpg')
        label_path = os.path.join(self.segmentation_class_path, f'{self.image_names[item]}.png')
        feature = self.read_image(feature_path, target_shape=self.target_shape,
                                  to_rgb=True, normalize=True)
        label = self.read_image(label_path, target_shape=self.target_shape,
                                to_rgb=True, normalize=False, color=(0, 0, 0))
        if self.transforms:
            feature, label = self.transforms(feature, label)
        else:
            feature, label = self.auguments([feature, label])
        if self.class_mapping:
            label = self.colormap2label(label, self.class_mapping)
        feature = self.hwc2chw(feature)
        encoded_label = self.hwc2chw(get_one_hot(label, num_classes=self.num_classes))

        return feature, encoded_label




if __name__ == "__main__":
    root = r'E:\PythonProject\Pytorch_Segmentation_Auorui\data\VOCdevkit\VOC2012'
    train_dataset = PascalVOCSegDataset(root, is_train=True, class_mapping=VOC_COLOR)
    # display_dataset_samples(voc_dataset, num_samples=3)
    val_dataset = PascalVOCSegDataset(root, is_train=False, class_mapping=VOC_COLOR)

    train_loader = DataLoader(train_dataset, shuffle=True, batch_size=2, num_workers=2, pin_memory=True)
    val_loader = DataLoader(val_dataset, batch_size=1, num_workers=2, pin_memory=True)
    print(len(train_dataset), len(val_dataset))
    # for feature, encoded_label in train_loader:
    #     print(feature.shape, encoded_label.shape)
    #     print(np.unique(encoded_label))



    def display_dataset_samples(dataset, num_samples=3):
        """
        显示数据集中的图像和标签。

        Args:
            dataset (PascalVOCSegDataset): 数据集对象。
            num_samples (int): 要显示的样本数量。
        """
        rows = num_samples
        cols = 2
        import matplotlib
        import matplotlib.pyplot as plt
        matplotlib.use("TkAgg")
        figure, ax = plt.subplots(nrows=rows, ncols=cols, figsize=(10, rows * 4))

        for i in range(num_samples):
            feature, encoded_label = dataset[i]

            image = np.transpose(feature, (1, 2, 0))  # CHW -> HWC
            label = np.argmax(encoded_label, axis=0)  # One-Hot -> 类别索引

            ax[i, 0].imshow(image)
            ax[i, 0].set_title("Image")
            ax[i, 0].set_axis_off()

            ax[i, 1].imshow(label, cmap="jet", interpolation="nearest")
            ax[i, 1].set_title("Ground Truth Mask")
            ax[i, 1].set_axis_off()

        plt.tight_layout()
        plt.show()


import cv2
import pyzjr.Z as Z

def AddText(img, text, x, y, color, thickness=1, font_scale=1,
            font_face=cv2.FONT_HERSHEY_SIMPLEX, line_type=cv2.LINE_AA, bottom_left_origin=False):
    """
    Adds text to an image at a specified location using OpenCV's `cv2.putText` method.

    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) to which the text will be added.
    text : str
        The text string to be drawn on the image.
    x : int
        The x-coordinate of the bottom-left corner of the text.
    y : int
        The y-coordinate of the bottom-left corner of the text.
    color : tuple
        The color of the text in BGR format (e.g., (255, 0, 0) for blue).
    thickness : int, optional
        The thickness of the text lines (default is 1).
    font_scale : float, optional
        The scale factor for the font size (default is 1).
    font_face : int, optional
        The font type (default is `cv2.FONT_HERSHEY_SIMPLEX`).
    line_type : int, optional
        The type of line used to draw the text (default is `cv2.LINE_AA` for anti-aliased lines).
    bottom_left_origin : bool, optional
        If True, the origin of the text is at the bottom-left corner of the image.
        If False, the origin is at the top-left corner (default is False).

    Usage Example:
    --------------
    >>> import numpy as np
    >>> img = np.zeros((500, 800, 3), dtype=np.uint8)
    >>> AddText(img, "Hello, OpenCV!", 50, 100, (0, 255, 0), 2)
    """
    cv2.putText(
        img,
        text,
        (x, y),
        fontFace=font_face,
        fontScale=font_scale,
        thickness=thickness,
        color=color,
        lineType=line_type,
        bottomLeftOrigin=bottom_left_origin
    )
    return img

def PutMultiLineText(img, text, org, color, thickness=1, font_scale=1,
                     font_face=cv2.FONT_HERSHEY_SIMPLEX, line_type=cv2.LINE_8, bottom_left_origin=False):
    """
    Draws multi-line text on an image at a specified location, with support for line breaks (`\n`).

    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) to which the text will be added.
    text : str
        The text string to be drawn on the image. Line breaks (`\n`) are supported.
    org : tuple
        The coordinates `(x, y)` of the bottom-left corner of the first line of text.
    color : tuple
        The color of the text in BGR format (e.g., (255, 0, 0) for blue).
    thickness : int, optional
        The thickness of the text lines (default is 1).
    font_scale : float, optional
        The scale factor for the font size (default is 1).
    font_face : int, optional
        The font type (default is `cv2.FONT_HERSHEY_SIMPLEX`).
    line_type : int, optional
        The type of line used to draw the text (default is `cv2.LINE_8`).
    bottom_left_origin : bool, optional
        If True, the origin of the text is at the bottom-left corner of the image.
        If False, the origin is at the top-left corner (default is False).

    Usage Example:
    --------------
    >>> import numpy as np
    >>> img = np.zeros((500, 800, 3), dtype=np.uint8)
    # >>> text = "Hello, OpenCV!\nThis is a test.\nCentered text with line breaks."
    >>> PutMultiLineText(img, text, (50, 50), (0, 255, 0), 1)
    """
    x, y = org
    text_lines = text.split('\n')
    # 获取文本行的高度（所有行的高度相同）
    _, line_height = cv2.getTextSize('', font_face, font_scale, thickness)[0]
    line_gap = line_height // 3

    for i, text_line in enumerate(text_lines):
        # 查找此行之前的文本的总大小
        line_y_adjustment = i * (line_gap + line_height)
        # 根据行号将文本从原始行向下移动
        if not bottom_left_origin:
            line_y = y + line_y_adjustment
        else:
            line_y = y - line_y_adjustment
        cv2.putText(img,
                    text=text_lines[i],
                    org=(x, line_y),
                    fontFace=font_face,
                    fontScale=font_scale,
                    color=color,
                    thickness=thickness,
                    lineType=line_type,
                    bottomLeftOrigin=bottom_left_origin)
    return img

def PutMultiLineCenteredText(img, text, color, thickness=1, font_scale=1,
                             font_face=cv2.FONT_HERSHEY_SIMPLEX, line_type=cv2.LINE_8):
    """
    Draws multi-line text centered both horizontally and vertically on an image.

    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) to which the text will be added.
    text : str
        The text string to be drawn on the image. Line breaks (`\n`) are supported.
    color : tuple
        The color of the text in BGR format (e.g., (255, 0, 0) for blue).
    thickness : int, optional
        The thickness of the text lines (default is 1).
    font_scale : float, optional
        The scale factor for the font size (default is 1).
    font_face : int, optional
        The font type (default is `cv2.FONT_HERSHEY_SIMPLEX`).
    line_type : int, optional
        The type of line used to draw the text (default is `cv2.LINE_8`).

    Usage Example:
    --------------
    >>> import numpy as np
    >>> img = np.zeros((500, 800, 3), dtype=np.uint8)  # Create a blank black image
    # >>> text = "Hello, OpenCV!\nThis is a test.\nCentered text with line breaks."
    >>> PutMultiLineCenteredText(img, text, (0, 255, 0), 2)  # Add green centered multi-line text
    """
    img_h, img_w = img.shape[:2]
    text_lines = text.split('\n')
    # 获取文本行的高度（所有行的高度相同）
    _, line_height = cv2.getTextSize('', font_face, font_scale, thickness)[0]
    line_gap = line_height // 3

    text_block_height = len(text_lines) * (line_height + line_gap)
    text_block_height -= line_gap

    for i, text_line in enumerate(text_lines):
        line_width, _ = cv2.getTextSize(text_line, font_face, font_scale, thickness)[0]
        x = (img_w - line_width) // 2
        y = (img_h + line_height) // 2
        # 查找此行之前的文本的总大小
        line_adjustment = i * (line_gap + line_height)
        y += line_adjustment - text_block_height // 2 + line_gap

        cv2.putText(img,
                    text=text_lines[i],
                    org=(x, y),
                    fontFace=font_face,
                    fontScale=font_scale,
                    color=color,
                    thickness=thickness,
                    lineType=line_type)
    return img

def PutBoxText(img, bbox, text, text_pos_mode=-1,
               bboxcolor=Z.green, textcolor=Z.red, thickness=1, bboxthickness=1,
               font_scale=1, font_face=cv2.FONT_HERSHEY_SIMPLEX, line_type=cv2.LINE_8):
    """
    Draws a bounding box and text on an image. The text position can be customized relative to the bounding box.

    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) on which the bounding box and text will be drawn.
    bbox : tuple
        The bounding box coordinates in the format `(x1, y1, x2, y2)`, where `(x1, y1)` is the top-left corner
        and `(x2, y2)` is the bottom-right corner.
    text : str
        The text string to be drawn on the image.
    text_pos_mode : int, optional
        Determines the position of the text relative to the bounding box:
        - `0`: Centered inside the bounding box.
        - `-1`: Top-left corner of the bounding box.
        - `1`: Top-right corner of the bounding box.
        Default is `0`.
    bboxcolor : tuple, optional
        The color of the bounding box in BGR format (e.g., (0, 255, 0) for green). Default is `Z.green`.
    textcolor : tuple, optional
        The color of the text in BGR format (e.g., (0, 0, 255) for red). Default is `Z.red`.
    thickness : int, optional
        The thickness of the text lines (default is 2).
    bboxthickness : int, optional
        The thickness of the bounding box lines (default is 2).
    font_scale : float, optional
        The scale factor for the font size (default is 1).
    font_face : int, optional
        The font type (default is `cv2.FONT_HERSHEY_SIMPLEX`).
    line_type : int, optional
        The type of line used to draw the text and bounding box (default is `cv2.LINE_8`).

    Usage Example:
    --------------
    >>> import numpy as np
    >>> img = np.zeros((500, 800, 3), dtype=np.uint8)
    >>> bbox = (150, 150, 300, 250)
    >>> PutBoxText(img, bbox, "Object", text_pos_mode=-1, bboxcolor=(0, 255, 0), textcolor=(0, 0, 255))
    """
    text_x = text_y = None
    x1, y1, x2, y2 = bbox
    text_size, _ = cv2.getTextSize(text, font_face, font_scale, thickness)
    cv2.rectangle(img, (x1, y1), (x2, y2), bboxcolor, bboxthickness)

    if text_pos_mode == 0:
        # Centered
        text_x = int((x1 + x2) / 2 - text_size[0] / 2)
        text_y = int((y1 + y2) / 2 + text_size[1] / 2)
    elif text_pos_mode == -1:
        # Top-left corner
        text_x = int(x1)
        text_y = int(y1 - text_size[1])
    elif text_pos_mode == 1:
        # Top-right corner
        text_x = int(x2 - text_size[0])
        text_y = int(y1 - text_size[1])

    cv2.putText(img,
                text=text,
                org=(text_x, text_y),
                fontFace=font_face,
                fontScale=font_scale,
                color=textcolor,
                thickness=thickness,
                lineType=line_type)
    return img

def PutRectangleText(img, text, bgcolor=(255, 255, 255), textcolor=(0, 0, 0), thickness=2,
                     font_scale=1.0, padding=10, font_face=cv2.FONT_HERSHEY_SIMPLEX,
                     line_type=cv2.LINE_8):
    """
    Draws a rectangle with a background color and places text inside it on an image.

    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) on which the rectangle and text will be drawn.
    text : str
        The text string to be drawn inside the rectangle.
    bgcolor : tuple, optional
        The background color of the rectangle in BGR format (e.g., (255, 255, 255) for white).
    textcolor : tuple, optional
        The color of the text in BGR format (e.g., (0, 0, 0) for black). Default is `Z.black`.
    thickness : int, optional
        The thickness of the text lines (default is 2).
    font_scale : float, optional
        The scale factor for the font size (default is 1.0).
    padding : int, optional
        The padding (in pixels) between the text and the edges of the rectangle (default is 10).
    font_face : int, optional
        The font type (default is `cv2.FONT_HERSHEY_SIMPLEX`).
    line_type : int, optional
        The type of line used to draw the text and rectangle (default is `cv2.LINE_8`).

    Usage Example:
    --------------
    >>> import numpy as np
    >>> img = np.zeros((500, 800, 3), dtype=np.uint8)  # Create a blank black image
    >>> PutRectangleText(img, "Hello, World!", bgcolor=(255, 255, 255), textcolor=(0, 0, 0))
    """
    (text_width, text_height), _ = cv2.getTextSize(text, font_face, font_scale, thickness)
    x1 = y1 = padding
    x2 = x1 + text_width + 2 * padding
    y2 = y1 + text_height + 2 * padding
    cv2.rectangle(img, (x1, y1), (x2, y2), bgcolor, -1)

    cv2.putText(img,
                text=text,
                org=(x1 + padding, y1 + text_height + padding),
                fontFace=font_face,
                fontScale=font_scale,
                color=textcolor,
                thickness=thickness,
                lineType=line_type)

    return img


if __name__=="__main__":
    import numpy as np
    img = np.zeros((500, 800, 3), dtype=np.uint8)
    text = "Hello, OpenCV!\nThis is a test.\nCentered text with line breaks."
    PutMultiLineText(img, text, (50, 50), (0, 255, 0), 1)
    PutMultiLineCenteredText(img, text, (0, 255, 255), 1)
    bbox = (150, 150, 300, 250)
    PutBoxText(img, bbox, "Object", text_pos_mode=-1, bboxcolor=(0, 255, 0), textcolor=(0, 0, 255))
    PutRectangleText(img, "Hello, World!", bgcolor=(255, 255, 255), textcolor=(0, 0, 0))
    cv2.imshow("test", img)
    cv2.waitKey(0)


"""
Copyright (c) 2024, Auorui.
All rights reserved.

Image enhancement suitable for OpenCV and Pillow
"""
import cv2
import torch
import numpy as np
import random
from math import ceil
from PIL import Image, ImageOps, ImageEnhance, ImageFilter
from pyzjr.utils.check import is_pil, is_numpy, get_image_size, get_image_num_channels, \
    is_positive_int, is_list_or_tuple
from pyzjr.augmentation.transforms.applytransformer import Images
from pyzjr.utils.randfun import rand

class ToTensor(object):
    """转换为Tensor格式"""
    def __init__(self, half=False):
        super(ToTensor, self).__init__()
        self.half = half

    def pil_to_tensor(self, pil_img):
        img = torch.as_tensor(np.array(pil_img, copy=True))
        img = img.view(pil_img.size[1], pil_img.size[0], len(pil_img.getbands()))
        img = img.permute((2, 0, 1))  # 从HWC格式转换为CHW格式
        return img

    def cv_to_tensor(self, np_img):
        np_img = np_img[:, :, ::-1]
        img = np.ascontiguousarray(np_img.transpose((2, 0, 1)))  # BGR to RGB -> HWC to CHW -> contiguous
        img = torch.from_numpy(img)
        return img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_to_tensor(image)
        elif is_numpy(image):
            image = self.cv_to_tensor(image)
        img = image.half() if self.half else image.float()  # 将uint8转换为fp16/32
        return img

class RandomHorizontalFlip(object):
    """该类用于随机水平翻转"""
    def __init__(self, pro=0.5):
        super(RandomHorizontalFlip, self).__init__()
        self.p = pro
        self.random_float = random.random()

    def pil_hflip(self, img):
        return img.transpose(Images.FLIP_LEFT_RIGHT)

    def cv_hflip(self, img):
        return cv2.flip(img, 1)

    def __call__(self, image):
        if self.random_float >= self.p:
            if is_pil(image):
                image = self.pil_hflip(image)
            elif is_numpy(image):
                image = self.cv_hflip(image)
        return image

class RandomVerticalFlip():
    """该类用于随机垂直翻转"""
    def __init__(self, pro=0.5):
        super(RandomVerticalFlip, self).__init__()
        self.p = pro
        self.random_float = random.random()

    def pil_vflip(self, img):
        return img.transpose(Images.FLIP_TOP_BOTTOM)

    def cv_vflip(self, img):
        return cv2.flip(img, 0)

    def __call__(self, image):
        if self.random_float >= self.p:
            if is_pil(image):
                image = self.pil_vflip(image)
            elif is_numpy(image):
                image = self.cv_vflip(image)
        return image

class AdjustBrightness(object):
    """该类用于调制亮度"""
    def __init__(self, brightness_factor):
        super(AdjustBrightness, self).__init__()
        self.brightness_factor = brightness_factor

    def pil_brightness(self, pil_img, brightness_factor):
        enhancer = ImageEnhance.Brightness(pil_img)
        pil_img = enhancer.enhance(brightness_factor)
        return pil_img

    def cv_brightness(self, np_img, brightness_factor):
        opencv_img = cv2.convertScaleAbs(np_img, alpha=brightness_factor, beta=0)
        return opencv_img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_brightness(image, self.brightness_factor)
        elif is_numpy(image):
            image = self.cv_brightness(image, self.brightness_factor)
        return image

class RandomAdjustBrightness(AdjustBrightness):
    """默认50%的概率使用调制图像亮度"""
    def __init__(self, factor, pro=0.5):
        super(RandomAdjustBrightness, self).__init__(factor)  # 修正super()参数，调用父类的__init__
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class AdjustGamma(object):
    """该类用于调整图像的伽玛值"""
    def __init__(self, gamma, gain=1):
        super(AdjustGamma, self).__init__()
        self.gamma = gamma
        self.gain = gain

    def pil_gamma(self, img, gamma, gain):
        gamma_table = [(255 + 1 - 1e-3) * gain * pow(x / 255., gamma) for x in range(256)]
        if len(img.split()) == 3:
            gamma_table = gamma_table * 3
            img = img.point(gamma_table)
        elif len(img.split()) == 1:
            img = img.point(gamma_table)
        return img

    def cv_gamma(self, img, gamma, gain):
        img_np = np.array(img)
        img_gamma_corrected = ((img_np / 255.0) ** gamma) * 255.0 * gain
        img_gamma_corrected = np.clip(img_gamma_corrected, 0, 255).astype(np.uint8)
        return img_gamma_corrected

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_gamma(image, self.gamma, self.gain)
        elif is_numpy(image):
            image = self.cv_gamma(image, self.gamma, self.gain)
        return image

class RandomAdjustGamma(AdjustGamma):
    """默认50%的概率使用调整图像的伽玛值"""
    def __init__(self, gamma, gain=1, pro=0.5):
        super(RandomAdjustGamma, self).__init__(gamma=gamma, gain=gain)  # 修正super()参数，调用父类的__init__
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class ToHSV(object):
    """该类用于对图像应用HSV增强"""
    def __init__(self, hue_range=(-0.1, 0.1), saturation_range=(0.6, 1.4), value_range=(0.7, 1.3)):
        super(ToHSV, self).__init__()
        self.hue_range = hue_range
        self.saturation_range = saturation_range
        self.value_range = value_range

    def _hsv(self, hsv_image):
        hue_factor = rand(self.hue_range[0], self.hue_range[1])
        saturation_factor = rand(self.saturation_range[0], self.saturation_range[1])
        value_factor = rand(self.value_range[0], self.value_range[1])

        hsv_image[:, :, 0] = np.clip(hsv_image[:, :, 0] * hue_factor, 0, 179)
        hsv_image[:, :, 1] = np.clip(hsv_image[:, :, 1] * saturation_factor, 0, 255)
        hsv_image[:, :, 2] = np.clip(hsv_image[:, :, 2] * value_factor, 0, 255)
        return hsv_image

    def __call__(self, image):
        if is_pil(image):
            hsv_image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2HSV)
            hsv_image = self._hsv(hsv_image)
            _image = Image.fromarray(cv2.cvtColor(hsv_image, cv2.COLOR_HSV2RGB))
        if is_numpy(image):
            hsv_image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
            hsv_image = self._hsv(hsv_image)
            _image = cv2.cvtColor(hsv_image, cv2.COLOR_HSV2BGR)
        return _image

class RandomRotation():
    """该类用于对图像进行随机旋转"""
    def __init__(self, degrees=None):
        super(RandomRotation, self).__init__()
        self.degrees = degrees

    def pil_rotate(self, img, angle):
        return img.rotate(angle)

    def cv_rotate(self, image, angle, fill_value=(0, 0, 0)):
        height, width = image.shape[:2]
        rotation_matrix = cv2.getRotationMatrix2D((width / 2, height / 2), angle, 1)
        rotated_image = cv2.warpAffine(image, rotation_matrix, (width, height), borderValue=fill_value)
        return rotated_image

    def __call__(self, image):
        h, w = get_image_size(image)
        is_squre = True if h == w else False
        if self.degrees is not None:
            angle = random.uniform(self.degrees[0], self.degrees[1])
        else:
            if is_squre:
                angle = random.choice([0, 90, 180, 270])
            else:
                angle = random.choice([0, 180])
        if is_pil(image):
            image = self.pil_rotate(image, angle)
        elif is_numpy(image):
            image = self.cv_rotate(image, angle)
        return image

class GaussianBlur():
    """该类用于对图像进行高斯模糊"""
    def __init__(self, ksize:int):
        self.ksize = ksize
        if self.ksize not in {1, 3, 5, 7, 9}:
            raise ValueError(f"Invalid ksize value {ksize}.The available values are 1, 3, 5, 7, 9")

    def cv_blur(self, np_img):
        sigma = 0.3 * ((self.ksize - 1) * 0.5 - 1) + 0.8
        np_img = cv2.GaussianBlur(np_img, ksize=(self.ksize, self.ksize), sigmaX=sigma)
        return np_img

    def pil_blur(self, pil_img):
        pillow_img = pil_img.filter(ImageFilter.GaussianBlur(radius=self.ksize))
        return pillow_img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_blur(image)
        elif is_numpy(image):
            image = self.cv_blur(image)
        return image

class Grayscale():
    """该类用于灰度化"""
    def __init__(self):
        super(Grayscale, self).__init__()

    def pil_to_gray(self, img):
        channels = get_image_num_channels(img)
        if channels == 3:
            img = img.convert('L')
            np_img = np.array(img, dtype=np.uint8)
            np_img = np.dstack([np_img, np_img, np_img])
            img = Image.fromarray(np_img, 'RGB')
        return img

    def cv_to_gray(self, img):
        channels = get_image_num_channels(img)
        if channels == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        return img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_to_gray(image)
        elif is_numpy(image):
            image = self.cv_to_gray(image)
        return image

class EqualizeHistogram():
    """该类用于对图像进行直方图均衡化"""
    def __init__(self, clahe=True):
        self.clahe = clahe

    def cv_equalize(self, img):
        yuv = cv2.cvtColor(img, cv2.COLOR_BGR2YUV)
        if self.clahe:
            c = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
            yuv[:, :, 0] = c.apply(yuv[:, :, 0])
        else:
            yuv[:, :, 0] = cv2.equalizeHist(yuv[:, :, 0])
        return cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR)

    def pil_equalize(self, img):
        return ImageOps.equalize(img)

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_equalize(image)
        elif is_numpy(image):
            image = self.cv_equalize(image)
        return image

class RandomEqualizeHistogram(EqualizeHistogram):
    """默认50%的概率使用直方图均衡化"""
    def __init__(self, pro=0.5, clahe=True):
        super(RandomEqualizeHistogram, self).__init__(clahe=clahe)
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class RandomLighting():
    """该类用于对图像进行随机光照调整"""
    def __init__(self, alpha):
        self.alpha = alpha

    def cv_lighting(self, image):
        alpha_b = np.random.normal(loc=0.0, scale=self.alpha)
        alpha_g = np.random.normal(loc=0.0, scale=self.alpha)
        alpha_r = np.random.normal(loc=0.0, scale=self.alpha)
        table = np.array([
            [55.46 * -0.5675, 4.794 * 0.7192, 1.148 * 0.4009],
            [55.46 * -0.5808, 4.794 * -0.0045, 1.148 * -0.8140],
            [55.46 * -0.5836, 4.794 * -0.6948, 1.148 * 0.4203]
        ])
        pca_b = table[2][0] * alpha_r + table[2][1] * alpha_g + table[2][2] * alpha_b
        pca_g = table[1][0] * alpha_r + table[1][1] * alpha_g + table[1][2] * alpha_b
        pca_r = table[0][0] * alpha_r + table[0][1] * alpha_g + table[0][2] * alpha_b
        img_arr = np.array(image).astype(np.float64)
        img_arr[:, :, 0] += pca_b
        img_arr[:, :, 1] += pca_g
        img_arr[:, :, 2] += pca_r
        img_arr = np.uint8(np.minimum(np.maximum(img_arr, 0), 255))

        return img_arr

    def pil_lighting(self, image):
        factor = 1.0 + random.uniform(-self.alpha, self.alpha)
        factor = max(0.1, min(factor, 10.0))
        enhancer = ImageEnhance.Brightness(image)
        adjusted_image = enhancer.enhance(factor)

        return adjusted_image

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_lighting(image)
        elif is_numpy(image):
            image = self.cv_lighting(image)
        return image

class Centerzoom():
    """该类用于中心缩放"""
    def __init__(self, factor):
        super().__init__()
        assert factor >= 1, "Zoom factor must be greater than or equal to 1"
        self.factor = factor

    def cv_centerzoom(self, image):
        h, w = image.shape[:2]
        h_ch, w_ch = ceil(h / self.factor), ceil(w / self.factor)
        h_top, w_top = (h - h_ch) // 2, (w - w_ch) // 2
        zoomed_img = cv2.resize(image[h_top: h_top + h_ch, w_top: w_top + w_ch], (w, h), interpolation=cv2.INTER_LINEAR)

        return zoomed_img

    def pil_centerzoom(self, image):
        width, height = image.size
        new_width, new_height = ceil(width / self.factor), ceil(height / self.factor)
        left = (width - new_width) // 2
        top = (height - new_height) // 2
        right = (width + new_width) // 2
        bottom = (height + new_height) // 2

        cropped_img = image.crop((left, top, right, bottom))
        zoomed_img = cropped_img.resize((width, height), Image.LANCZOS)

        return zoomed_img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_centerzoom(image)
        elif is_numpy(image):
            image = self.cv_centerzoom(image)
        return image

class Resize(object):
    """该类用于调整图像大小到指定的尺寸"""
    def __init__(self, target_size):
        super(Resize, self).__init__()
        if is_positive_int(target_size):
            self.h = self.w = target_size
        elif is_list_or_tuple(target_size) and len(target_size) == 2:
            self.w, self.h = target_size
        self.target_size = target_size

    def pil_resize(self, img, interpolation=Image.BILINEAR):
        img_resized = img.resize((self.w, self.h), Images.LINEAR)
        return img_resized

    def cv_resize(self, img):
        image = cv2.resize(img, (self.w, self.h), interpolation=cv2.INTER_CUBIC)
        return image

    def __call__(self, img):
        if is_pil(img):
            img = self.pil_resize(img)
        elif is_numpy(img):
            img = self.cv_resize(img)
        return img

class RandomCrop(object):
    """该类用于随机裁剪出一个指定大小的区域。"""
    def __init__(self, target_size):
        super(RandomCrop, self).__init__()
        if is_positive_int(target_size):
            self.h = self.w = target_size
        elif is_list_or_tuple(target_size) and len(target_size) == 2:
            self.w, self.h = target_size
        self.target_size = target_size

    def pil_random_crop(self, img):
        width, height = img.size
        if width < self.w or height < self.h:
            raise ValueError("Image size should be larger than target size.")

        left = random.randint(0, width - self.w)
        top = random.randint(0, height - self.h)
        right = left + self.w
        bottom = top + self.h

        cropped_img = img.crop((left, top, right, bottom))
        return cropped_img

    def cv_random_crop(self, img):
        h, w = img.shape[:2]
        if w < self.w or h < self.h:
            raise ValueError("Image size should be larger than target size.")
        x = random.randint(0, w - self.w)
        y = random.randint(0, h - self.h)
        cropped_img = img[y:y+self.h, x:x+self.w]
        return cropped_img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_random_crop(image)
        elif is_numpy(image):
            image = self.cv_random_crop(image)
        return image

class RandomResizeCrop(RandomCrop):
    """类用于随机调整图像大小并裁剪出一个指定大小的区域。"""
    def __init__(self, target_size, scale_range=(1., 2.)):
        super(RandomResizeCrop, self).__init__(target_size=target_size)
        if scale_range[0] < 1.0 or scale_range[1] < 1.0:
            raise ValueError("Scale range must have values greater than or equal to 1.0")
        if scale_range[0] > scale_range[1]:
            raise ValueError("The lower bound of scale_range must be smaller than the upper bound")
        self.scale_range = scale_range
        self.scale_factor = random.uniform(self.scale_range[0], self.scale_range[1])

    def pil_random_resize(self, img):
        width, height = img.size
        new_width = int(width * self.scale_factor)
        new_height = int(height * self.scale_factor)
        resized_img = img.resize((new_width, new_height), Image.LANCZOS)
        return resized_img

    def cv_random_resize(self, img):
        h, w = img.shape[:2]
        new_w = int(w * self.scale_factor)
        new_h = int(h * self.scale_factor)
        resized_img = cv2.resize(img, (new_w, new_h), interpolation=cv2.INTER_CUBIC)
        return resized_img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_random_resize(image)
            image = self.pil_random_crop(image)
        if is_numpy(image):
            image = self.cv_random_resize(image)
            image = self.cv_random_crop(image)
        return image

class CenterCrop():
    """该类用于从图像中心裁剪出一个指定大小的区域。"""
    def __init__(self, target_size):
        super(CenterCrop, self).__init__()
        if is_positive_int(target_size):
            self.h = self.w = target_size
        elif is_list_or_tuple(target_size) and len(target_size) == 2:
            self.w, self.h = target_size
        self.target_size = target_size

    def pil_center_crop(self, img):
        img_width, img_height = img.size

        left = (img_width - self.w) // 2
        top = (img_height - self.h) // 2
        right = left + self.w
        bottom = top + self.h

        return img.crop((left, top, right, bottom))

    def cv_center_crop(self, img):
        img_height, img_width = img.shape[:2]

        start_x = (img_width - self.w) // 2
        start_y = (img_height - self.h) // 2
        end_x = start_x + self.w
        end_y = start_y + self.h

        return img[start_y:end_y, start_x:end_x]

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_center_crop(image)
        elif is_numpy(image):
            image = self.cv_center_crop(image)
        return image

class Pad():
    """该类用于对图像进行边缘填充。"""
    def __init__(self, padding, fill_value=(128, 128, 128)):
        super(Pad, self).__init__()
        if is_positive_int(padding):
            top = bottom = left = right = padding

        elif is_list_or_tuple(padding):
            if len(padding) == 2:
                left = right = padding[0]
                top = bottom = padding[1]
            elif len(padding) == 4:
                left, top, right, bottom = padding
            else:
                raise ValueError("The size of the padding list or tuple should be 2 or 4.")
        else:
            raise TypeError("Padding can be any of: a number, a tuple or list of size 2 or 4.")
        self.paddings = left, top, right, bottom
        self.fill_value = fill_value

    def pil_pad(self, img):
        left, top, right, bottom = self.paddings
        img_padded = ImageOps.expand(img, (left, top, right, bottom), self.fill_value)
        return img_padded

    def cv_pad(self, img):
        left, top, right, bottom = self.paddings
        img_padded = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=self.fill_value)
        return img_padded

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_pad(image)
        elif is_numpy(image):
            image = self.cv_pad(image)
        return image

class ResizePad():
    """
    该类用于调整图像大小并进行填充。
    提供了两种图像处理模式：调整大小后填充（frame=True）或直接调整大小（frame=False）。
    """
    def __init__(self, target_size, frame=True):
        super(ResizePad, self).__init__()
        if is_positive_int(target_size):
            self.h = self.w = target_size
        elif is_list_or_tuple(target_size) and len(target_size) == 2:
            self.w, self.h = target_size
        self.frame = frame

    def pil_resizepad(self, image):
        w, h = self.w, self.h
        iw, ih = image.size
        if self.frame:
            scale = min(w / iw, h / ih)
            nw = int(iw * scale)
            nh = int(ih * scale)
            image = image.resize((nw, nh), Image.BICUBIC)
            new_image = Image.new('RGB', (w, h), (128, 128, 128))
            new_image.paste(image, ((w-nw) // 2, (h-nh) // 2))
            return new_image
        else:
            return image.resize((w, h), Image.BICUBIC)

    def cv_resizepad(self, image):
        ih, iw = image.shape[:2]
        w, h = self.w, self.h
        if self.frame:
            scale = min(w / iw, h / ih)
            nw = int(iw * scale)
            nh = int(ih * scale)
            resized_image = cv2.resize(image, (nw, nh), interpolation=cv2.INTER_CUBIC)
            new_image = np.full((h, w, 3), (128, 128, 128), dtype=np.uint8)
            top = (h - nh) // 2
            left = (w - nw) // 2
            new_image[top:top + nh, left:left + nw] = resized_image
            return new_image
        else:
            return cv2.resize(image, (w, h), interpolation=cv2.INTER_CUBIC)

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_resizepad(image)
        if is_numpy(image):
            image = self.cv_resizepad(image)
        return image

class InvertColor():
    """该类用于对图像进行颜色反转，即将图像的像素颜色值从原色反转为补色。"""
    def __init__(self):
        super(InvertColor, self).__init__()

    def pil_invert_color(self, img):
        return ImageOps.invert(img)

    def cv_invert_color(self, img):
        return 255 - img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_invert_color(image)
        elif is_numpy(image):
            image = self.cv_invert_color(image)
        return image

class RandomInvertColor(InvertColor):
    """默认50%概率使用对图像进行颜色反转"""
    def __init__(self, pro=0.5):
        super(RandomInvertColor, self).__init__()
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class AdjustSharpness():
    """
    该类用于调整图像的锐度。锐度调整可以增强图像的边缘细节，使图像看起来更加清晰或更具纹理。
    锐度因子，> 1表示增加锐度，< 1表示降低锐度。
    """
    def __init__(self, factor):
        super(AdjustSharpness, self).__init__()
        self.sharpness_factor = factor

    def pil_adjust_sharpness(self, img):
        enhancer = ImageEnhance.Sharpness(img)
        img = enhancer.enhance(self.sharpness_factor)
        return img

    def cv_adjust_sharpness(self, img):
        kernel = np.array([[0, -1, 0],
                           [-1,  5, -1],
                           [0, -1, 0]])
        # kernel = kernel * self.sharpness_factor
        sharpened_img = cv2.filter2D(img, -1, kernel)
        return sharpened_img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_adjust_sharpness(image)
        elif is_numpy(image):
            image = self.cv_adjust_sharpness(image)
        return image

class RandomAdjustSharpness(AdjustSharpness):
    """默认50%概率使用调整图像的锐度"""
    def __init__(self, factor, pro=0.5):
        super(RandomAdjustSharpness, self).__init__(factor)  # 修正super()参数，调用父类的__init__
        self.sharpness_factor = factor
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class AdjustSaturation():
    """
    该类用于调整图像的饱和度，可以通过指定一个饱和度因子来增强或降低图像的饱和度。
    饱和度影响图像中色彩的浓淡，值越高，图像的颜色越鲜艳；值越低，图像的颜色越灰。
    """
    def __init__(self, factor):
        super(AdjustSaturation, self).__init__()
        self.saturation_factor = factor

    def pil_saturation(self, img):
        enhancer = ImageEnhance.Color(img)
        img = enhancer.enhance(self.saturation_factor)
        return img

    def cv_saturation(self, img):
        hsv_img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        hsv_img = np.array(hsv_img, dtype=np.float32)
        hsv_img[:, :, 1] = hsv_img[:, :, 1] * self.saturation_factor
        hsv_img[:, :, 1] = np.clip(hsv_img[:, :, 1], 0, 255)
        img = cv2.cvtColor(hsv_img.astype(np.uint8), cv2.COLOR_HSV2BGR)
        return img

    def __call__(self, image):
        if is_pil(image):
            image = self.pil_saturation(image)
        elif is_numpy(image):
            image = self.cv_saturation(image)
        return image

class RandomAdjustSaturation(AdjustSaturation):
    """默认50%概率使用调整图像的饱和度"""
    def __init__(self, factor, pro=0.5):
        super(RandomAdjustSaturation, self).__init__(factor)  # 修正super()参数，调用父类的__init__
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class ColorJitter():
    """
    该类用于对图像进行随机颜色变化处理，包括亮度、对比度、饱和度和色相的随机扰动。
    这些操作常用于数据增强，以增加训练数据的多样性，从而提高模型的泛化能力。
    """
    def __init__(self,
                 brightness_factor_range=(0.5, 1.5),  # 亮度范围
                 contrast_factor_range=(0.5, 1.5),    # 对比度范围
                 saturation_factor_range=(0.5, 1.5),  # 饱和度范围
                 hue_factor_range=(-0.1, 0.1)         # 色相范围
                 ):
        super(ColorJitter, self).__init__()
        self.brightness_factor_range = brightness_factor_range
        self.contrast_factor_range = contrast_factor_range
        self.saturation_factor_range = saturation_factor_range
        self.hue_factor_range = hue_factor_range

    def brightness(self, pil_img, brightness_factor):
        enhancer = ImageEnhance.Brightness(pil_img)
        pil_img = enhancer.enhance(brightness_factor)
        return pil_img

    def contrast(self, pil_img, contrast_factor):
        enhancer = ImageEnhance.Contrast(pil_img)
        img = enhancer.enhance(contrast_factor)
        return img

    def saturation(self, pil_img, saturation_factor):
        enhancer = ImageEnhance.Color(pil_img)
        img = enhancer.enhance(saturation_factor)
        return img

    def hue(self, pil_img, hue_factor):
        image = pil_img
        image_hue_factor = hue_factor
        if not -0.5 <= image_hue_factor <= 0.5:
            raise ValueError('image_hue_factor {} is not in [-0.5, 0.5].'.format(image_hue_factor))
        mode = image.mode
        if mode in {'L', '1', 'I', 'F'}:
            return image
        hue, saturation, value = pil_img.convert('HSV').split()
        np_hue = np.array(hue, dtype=np.uint8)

        # Correct hue wrapping
        np_hue = (np_hue + np.uint8(image_hue_factor * 255)) % 256

        hue = Image.fromarray(np_hue, 'L')
        image = Image.merge('HSV', (hue, saturation, value)).convert(mode)
        return image

    def _jitter(self, pil_img):
        brightness_factor = random.uniform(*self.brightness_factor_range)
        contrast_factor = random.uniform(*self.contrast_factor_range)
        saturation_factor = random.uniform(*self.saturation_factor_range)
        hue_factor = random.uniform(*self.hue_factor_range)
        # Apply jitter in the correct order
        img = self.brightness(pil_img, brightness_factor)
        img = self.contrast(img, contrast_factor)
        img = self.saturation(img, saturation_factor)
        img = self.hue(img, hue_factor)
        return img

    def __call__(self, image):
        if is_numpy(image):
            image = Image.fromarray(image[:, :, ::-1])
            image = self._jitter(image)
            image = np.array(image)[:, :, ::-1]
        elif is_pil(image):
            image = self._jitter(image)
        return image

class MeanStdNormalize():
    """ 应放在 ToTensor 之后
    该类用于对图像进行基于均值（mean）和标准差（std）的归一化操作。
    如果没有提供均值和标准差，它将默认使用ImageNet数据集的常见均值和标准差进行归一化。
    """
    def __init__(self, mean=None, std=None):
        super(MeanStdNormalize, self).__init__()
        self.IMAGENET_MEAN = 0.485, 0.456, 0.406
        self.IMAGENET_STD = 0.229, 0.224, 0.225
        self.mean = self.IMAGENET_MEAN if mean is None else mean
        self.std = self.IMAGENET_STD if std is None else std

    def mean_std_normalize(self, img):
        if img.ndim < 3:
            raise ValueError(f'Expected tensor to be a tensor image of size (..., C, H, W). Got tensor.size() = {img.size()}.')

        mean = torch.as_tensor(self.mean, dtype=img.dtype, device=img.device)
        std = torch.as_tensor(self.std, dtype=img.dtype, device=img.device)
        if (std == 0).any():
            raise ValueError('std evaluated to zero after conversion to {}, leading to division by zero.'.format(img.dtype))
        if mean.ndim == 1:
            mean = mean.view(-1, 1, 1)
        if std.ndim == 1:
            std = std.view(-1, 1, 1)
        img.sub_(mean).div_(std)
        return img

    def __call__(self, img):
        return self.mean_std_normalize(img)

if __name__=="__main__":
    import matplotlib
    import matplotlib.pyplot as plt
    from pyzjr.augmentation.transforms.applytransformer import Compose
    matplotlib.use("TkAgg")

    image_path = r'../../data/scripts/fire.jpg'
    cv_image = cv2.imread(image_path)
    pil_image = Image.open(image_path)

    transforms = Compose([
        # RandomHorizontalFlip(),
        # RandomVerticalFlip(),
        # AdjustBrightness(0.8),
        # RandomAdjustBrightness(1.2),
        # AdjustGamma(1.2),
        # RandomAdjustGamma(1.3),
        # ToHSV(),
        # RandomRotation(),
        # GaussianBlur(ksize=3),
        # Grayscale(),
        # EqualizeHistogram(),
        # RandomEqualizeHistogram(),
        # Centerzoom(1.6),
        # Resize(target_size=(256, 256)),
        # RandomCrop(target_size=128),
        # CenterCrop(target_size=128),
        # RandomResizeCrop(target_size=128),
        # Pad((40, 15)),
        # ResizePad(256),
        # InvertColor(),
        # RandomInvertColor(),
        # AdjustSharpness(1.2),
        # RandomAdjustSharpness(1.5),
        # AdjustSaturation(3),
        # RandomAdjustSaturation(3),
        # ColorJitter(),
        # Normalizer('zero_centered'),
        ToTensor(),
        MeanStdNormalize()
    ])
    transformed_pil_image = transforms(pil_image)
    transformed_cv_image = transforms(cv_image)

    print(transformed_pil_image.shape)
    print(transformed_cv_image.shape)
    transformed_pil_image = torch.clamp(transformed_pil_image, 0, 1)
    transformed_pil_image = transformed_pil_image.permute(1, 2, 0)

    transformed_cv_image = torch.clamp(transformed_cv_image, 0, 1)
    transformed_cv_image = transformed_cv_image.permute(1, 2, 0)

    plt.figure(figsize=(6, 6))
    plt.subplot(1, 2, 1)
    plt.imshow(transformed_pil_image)
    plt.title('Transformed PIL Image')
    plt.axis('off')

    plt.subplot(1, 2, 2)
    plt.imshow(transformed_cv_image)
    plt.title('Transformed CV Image')
    plt.axis('off')

    plt.show()



import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["DropPath", "Dropout", "MultiSampleDropout", "DropConnect",
           "Standout", "GaussianDropout"]

class DropPath(nn.Module):
    """
    Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).

    I follow the implementation:
        https://github.com/rwightman/pytorch-image-models/blob/a2727c1bf78ba0d7b5727f5f95e37fb7f8866b1f/timm/models/layers/drop.py
    """
    def __init__(self, drop_prob = 0.):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        """
        DropPath类实现了"Drop Path"的正则化技术，用于深度神经网络中的主路径，特别是在残差块的主路径上。Drop Path类似于Dropout，
        但它是在网络的深度路径上随机丢弃整个路径（或称之为跳跃连接）。这有助于训练更加健壮和泛化性能更好的神经网络。
        """
        return self.drop_path(x, self.drop_prob, self.training)

    def __str__(self):
        return f'drop_prob={round(self.drop_prob,3):0.3f}'

    def drop_path(self, x, drop_prob=0., training=False):
        """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).
        This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
        the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
        See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
        changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
        'survival rate' as the argument.
        """
        if drop_prob == 0. or not training:
            return x
        keep_prob = 1 - drop_prob
        shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
        random_tensor = keep_prob + torch.rand(
            shape, dtype=x.dtype, device=x.device)
        output = x.div(keep_prob) * random_tensor.floor()
        return output


class Dropout(nn.Module):
    """This module provides a customizable dropout layer. Dropout is a regularization
    technique commonly used in neural networks during training to prevent overfitting.
    It randomly sets a fraction of input units to zero during each update, which helps
    prevent co-adaptation of hidden units.

    Reference from paper: https://arxiv.org/abs/1207.0580
    """
    def __init__(self, p=.5, training=True, inplace=True):
        super().__init__()
        self.p = p
        self.training = training
        self.inplace = inplace

    def forward(self, input):
        return F.dropout(input, self.p, self.training, self.inplace)


class MultiSampleDropout(nn.Module):
    """The core idea is to apply Dropout multiple times to the same input in a single
    forward propagation process, generate multiple different masks, and aggregate the
    results.

    Reference from paper: https://arxiv.org/pdf/1905.09788.pdf"""
    def __init__(self, dropout_rate, num_samples):
        super(MultiSampleDropout, self).__init__()
        self.dropout_rate = dropout_rate
        self.num_samples = num_samples

    def forward(self, x):
        outputs = []
        for _ in range(self.num_samples):
            output = F.dropout(x, self.dropout_rate, training=self.training)
            outputs.append(output)
        return torch.mean(torch.stack(outputs), dim=0)


class DropConnect(nn.Module):
    """Dropout works by randomly setting the activation output of neurons to zero,
    while DropConnect randomly sets the weights of the network to zero. This means
    that in DropConnect, the connection (i.e. weight) portion of the network is
    randomly "discarded" rather than outputted.

    Reference from paper: https://proceedings.mlr.press/v28/wan13.pdf"""
    def __init__(self, input_dim, output_dim, drop_prob=0.5, training=True):
        super(DropConnect, self).__init__()
        self.drop_prob = drop_prob
        self.weight = nn.Parameter(torch.randn(input_dim, output_dim))
        self.bias = nn.Parameter(torch.randn(output_dim))
        self.training = training

    def forward(self, x):
        if self.training:       # 生成与权重相同形状的掩码, 用掩码乘以权重
            mask = torch.rand(self.weight.size()) > self.drop_prob
            drop_weight = self.weight * mask.float().to(self.weight.device)
        else:       # 在测试时不应用DropConnect，但要调整权重以反映丢弃率
            drop_weight = self.weight * (1 - self.drop_prob)
        return F.linear(x, drop_weight, self.bias)


class Standout(nn.Module):
    """Using the internal state of neural networks to determine which neurons are
    more likely to be retained, making the regularization process more dependent on
    the current behavior of the model. This adaptability enables Standout to achieve
    personalized regularization intensity at different training stages and data points.
    """
    def __init__(self, input_dim, output_dim, training=True):
        super(Standout, self).__init__()
        self.fc = nn.Linear(input_dim, output_dim)
        self.prob_fc = nn.Linear(input_dim, output_dim)   # 使用一个简单的全连接层来生成保留概率
        self.training = training

    def forward(self, x):
        #Note: Standout should not be applied in evaluation mode
        x_out = self.fc(x)
        if self.training:
            probs = torch.sigmoid(self.prob_fc(x))      # 计算每个神经元的保留概率
            mask = torch.bernoulli(probs).to(x.device)  # 生成与激活大小相同的二值掩码
            x_out = x_out * mask                        # 应用mask
        return x_out


class GaussianDropout(nn.Module):
    """
    When training neural networks, overfitting is prevented by introducing random
    noise that follows a Gaussian distribution. Unlike traditional Dropout, which
    sets activation to zero with a fixed probability, Gaussian Dropout is multiplied
    by a random variable (m), where (m) follows a Gaussian distribution.
    """
    def __init__(self, sigma=0.1, training=True):
        super(GaussianDropout, self).__init__()
        self.sigma = sigma
        self.training = training

    def forward(self, x):
        # Note: Dropout is not used during the testing phase
        if self.training:
            m_with_gaussian_noise = torch.normal(1.0, self.sigma, size=x.size()).to(x.device)
            return x * m_with_gaussian_noise
        else:
            return x


"""
Copyright (c) 2023, Auorui.
All rights reserved.
"""
import cv2
import numpy as np
from pyzjr.augmentation.mask.contour import check_points_in_contour
from pyzjr.utils.check import is_gray_image

def unique(image):
    """Returns a list of unique pixel values in the input image"""
    np_image = np.array(image).astype(np.uint8)
    return np.unique(np_image)

def count_zero(thresh):
    """计算矩阵0值"""
    zero_count = np.sum(thresh == 0)
    return zero_count

def count_white(thresh):
    """计算矩阵255值"""
    white_count = np.sum(thresh == 255)
    return white_count

def count_nonzero(thresh):
    """计算矩阵非0值"""
    nonzero_pixels = np.count_nonzero(thresh)
    return nonzero_pixels

def incircleV1(image, contours_arr, is_draw=True, color=(0, 0, 255), use_check_point=False):
    """
    计算并绘制轮廓的最大内切圆。
    该方法通过对每个轮廓计算轮廓内部的最大内切圆，并返回最大内切圆的直径和位置。

    :param image: 单通道图像，通常是二值化后的图像，用于计算最大内切圆。
    :param contours_arr: 轮廓列表，通常通过 `cv2.findContours` 获得。每个轮廓是一个二维点集。
    :param is_draw: 是否在结果图像上绘制最大内切圆，默认为 True。
                     如果为 True，则在图像上绘制内切圆。
    :param color: 绘制内切圆的颜色，默认为红色 (0, 0, 255)。
    :param use_check_point: 是否进行内切圆验证，通过检查圆心及其半径内的点是否都在轮廓内部。
                             该参数主要适用于裂缝类型为网状结构时使用，默认值为 False。
    :return: 如果 `is_draw=True`，返回包含绘制内切圆的图像以及内切圆的直径。
             如果 `is_draw=False`，仅返回内切圆的直径。
    """
    if is_gray_image(image):
        result = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
    else:
        result = image.copy()

    max_val = 0
    max_dist_pt = None
    # 只考虑非零点
    non_zero_points = np.array(np.nonzero(image))
    non_zero_points = non_zero_points.T
    for contours in contours_arr:
        raw_dist = np.zeros(image.shape[:2], dtype=np.float32)
        # 使用cv2.pointPolygonTest计算每个非零点到轮廓的距离
        for point in non_zero_points:
            x, y = point
            x, y = int(x), int(y)
            raw_dist[x, y] = cv2.pointPolygonTest(contours, (y, x), True)
        # for i in range(image.shape[0]):
        #     for j in range(image.shape[1]):
        #         raw_dist[i, j] = cv2.pointPolygonTest(contours, (j, i), True)
        # 寻找最大内切圆直径及其圆心
        min_val, curr_max_val, _, curr_max_dist_pt = cv2.minMaxLoc(raw_dist)
        # 确保选中的点在轮廓内部
        if curr_max_val > max_val:
            # 检查圆心加上半径后的八个方向点是否都在轮廓内部
            if use_check_point:
                if check_points_in_contour(curr_max_dist_pt, int(curr_max_val)-2, contours):
                    max_val = curr_max_val
                    max_dist_pt = curr_max_dist_pt
            else:
                max_val = curr_max_val
                max_dist_pt = curr_max_dist_pt
    wide = max_val * 2
    if is_draw and max_dist_pt is not None:
        result = cv2.circle(result, max_dist_pt, int(max_val), color, 1, 1, 0)

    if is_draw:
        return wide, result
    else:
        return wide

def incircleV2(image, is_draw=True, color=(0, 0, 255)):
    """
    计算并绘制图像的最大内切圆。
    该方法通过距离变换 (`cv2.distanceTransform`) 来计算图像的最大内切圆的直径和位置。

    :param image: 单通道图像，通常是二值化后的图像，用于计算最大内切圆。
    :param is_draw: 是否在结果图像上绘制最大内切圆，默认为 True。
                     如果为 True，则在图像上绘制内切圆。
    :param color: 绘制内切圆的颜色，默认为红色 (0, 0, 255)。
    :return: 如果 `is_draw=True`，返回包含绘制内切圆的图像以及内切圆的直径。
             如果 `is_draw=False`，仅返回内切圆的直径。
    """
    if is_gray_image(image):
        result = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
    else:
        result = image.copy()

    max_val = 0
    max_dist_pt = None
    dist_transform = cv2.distanceTransform(image, cv2.DIST_L2, 3)
    min_val, curr_max_val, _, curr_max_dist_pt = cv2.minMaxLoc(dist_transform)
    if curr_max_val > max_val:
        max_val = curr_max_val
        max_dist_pt = curr_max_dist_pt
    wide = max_val * 2
    if is_draw and max_dist_pt is not None:
        result = cv2.circle(result, max_dist_pt, int(max_val), color, 1, 1, 0)
    if is_draw:
        return wide, result
    else:
        return wide

def outcircle(img, contours_arr, is_draw=True, color=(0, 255, 0)):
    """
    轮廓外切圆算法,画出所有轮廓的外切圆
    :param img: 单通道图像
    :param contours_arr: ndarry的轮廓, 建议使用cv2.findContours,
                        contours_arr, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
                        pyzjr也有SearchOutline可用,但要转换成ndarry的格式
    :example:
        contour = [np.array([point], dtype=np.int32) for point in contours]
    :param color: 绘制外切圆颜色
    :return: 包含结果图像（如果is_draw为True）和外切圆信息（所有轮廓的外切圆圆心和半径）的元组。
           如果is_draw为False，则只返回外切圆信息。
    """
    radii = []
    if is_draw:
        result = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    else:
        result = None

    for cnt in contours_arr:
        (x, y), radius = cv2.minEnclosingCircle(cnt)
        radii.append([(x, y), radius])
        if is_draw:
            center = (int(x), int(y))
            radius = int(radius)
            cv2.circle(result, center, radius, color, 1)  # 绘制外切圆
            cv2.circle(result, center, 1, color, 1)       # 绘制圆心

    if is_draw:
        return radii, result
    else:
        return radii


if __name__ == "__main__":
    from pyzjr.augmentation.mask.predeal import binarization
    image = cv2.imread(r"E:\PythonProject\pyzjrPyPi\pyzjr\measure\fissure\0003.png")
    thresh = binarization(image)
    # contours_arr = SearchOutline(thresh)
    contours_arr, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    wide, result = incircleV1(thresh, contours_arr)
    print(wide)
    wide, result_transform = incircleV2(thresh)
    print(wide)
    # radii, results = outcircle(thresh, contours_arr)
    # print(wide, radii)
    # cv2.imwrite("sss.png", result_transform)
    cv2.imshow("ss", result_transform)
    cv2.waitKey(0)



"""
Copyright (c) 2022, Auorui.
All rights reserved.
"""
import cv2
import numpy as np
from pathlib import Path
from urllib import request

from pyzjr.utils.check import is_file, is_gray_image, is_numpy, is_list, is_url, is_pil
from pyzjr.utils.randfun import randstring

def bgr_read(filename, is_gray=False, flip=None, to_rgb=False):
    """
    自定义的opencv读取函数
    :param filename: Path to the image file.
    :param is_gray: If True, read the image in grayscale.
    :param flip: 0: Flip vertically.
                1: Flip horizontally.
               -1: Flip both vertically and horizontally.
    :param to_rgb: Convert to RGB, default to False

    Examples:
    ```
        image_path = r'dog.png'
        gray_image = bgr_read(image_path, is_gray=True, to_rgb=True)
        display("ss", gray_image)
        imwriter("ss.png", gray_image)
    ```
    """
    if not is_file(filename):
        raise ValueError(f"The file {filename} does not exist.")
    if isinstance(filename, Path):
        filename = str(filename.resolve())
    flags = cv2.IMREAD_GRAYSCALE if is_gray else cv2.IMREAD_COLOR
    image = cv2.imdecode(np.fromfile(filename, np.uint8), flags)
    if flip is not None:
        assert flip in [-1, 0, 1], f"The 'flip' parameter must be -1, 0, or 1."
        image = cv2.flip(image, flip)
    if is_gray_image(image):
        image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB) if to_rgb else image
    else:
        image = cv2.cvtColor(image, cv2.COLOR_BAYER_BG2RGB) if to_rgb else image
    return image

def imwriter(filename: str, img: np.ndarray, params=None):
    """Write the image to a file."""
    try:
        cv2.imencode(Path(filename).suffix, img, params)[1].tofile(filename)
        return True
    except Exception:
        return False

def display(imgArray, winname=None, scale=1.):
    """Displays an image in the specified window."""
    _imshow = cv2.imshow  # copy to avoid recursion errors
    if winname is None:
        winname = randstring(5)
    if is_numpy(imgArray):
        height, width = imgArray.shape[:2]
        new_width = int(width * scale)
        new_height = int(height * scale)
        image = cv2.resize(imgArray, (new_width, new_height))
        _imshow(winname.encode('unicode_escape').decode(), image)
    elif is_list(imgArray):
        image = StackedImagesV1(scale, imgArray)
        _imshow(winname.encode('unicode_escape').decode(), image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()

def url2image(url):
    """Same usage as cv2.imread()"""
    if is_url(url):
        res = request.urlopen(url, timeout=3)
    else:
        raise ValueError("The current input parameter does not conform to the URL format")
    try:
        image = np.asarray(bytearray(res.read()), dtype="uint8")
        image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
    except:
        print('Load read - Image timeout!')
        image = []
    h, w, c = image.shape
    if c == 4:
        image = image[:, :, :3]
    return image

def stacked2image(image1, image2, is_hstack=True):
    """Simply overlaying two images"""
    stack = np.hstack if is_hstack else np.vstack
    image = stack([image1, image2])
    return image

def StackedImagesV1(scale, imgArray):
    """
    Display Images According to List Structure

    :param scale: The scale of the images, where 1 represents the original size.
    :param imgArray: A list of images representing the arrangement in rows and columns.
    :return: A generated image that displays the images in the order specified by the input list, arranged in a grid.
    """
    rows = len(imgArray)
    cols = len(imgArray[0])
    rowsAvailable = isinstance(imgArray[0], list)
    width = imgArray[0][0].shape[1]
    height = imgArray[0][0].shape[0]
    if rowsAvailable:
        for x in range(0, rows):
            for y in range(0, cols):
                if imgArray[x][y].shape[:2] == imgArray[0][0].shape [:2]:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (0, 0), None, scale, scale)
                else:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (imgArray[0][0].shape[1], imgArray[0][0].shape[0]), None, scale, scale)
                if len(imgArray[x][y].shape) == 2: imgArray[x][y]= cv2.cvtColor( imgArray[x][y], cv2.COLOR_GRAY2BGR)
        imageBlank = np.zeros((height, width, 3), np.uint8)
        hor = [imageBlank] * rows
        for x in range(0, rows):
            hor[x] = np.hstack(imgArray[x])
        ver = np.vstack(hor)
    else:
        for x in range(0, rows):
            if imgArray[x].shape[:2] == imgArray[0].shape[:2]:
                imgArray[x] = cv2.resize(imgArray[x], (0, 0), None, scale, scale)
            else:
                imgArray[x] = cv2.resize(imgArray[x], (imgArray[0].shape[1], imgArray[0].shape[0]), None,scale, scale)
            if len(imgArray[x].shape) == 2:
                imgArray[x] = cv2.cvtColor(imgArray[x], cv2.COLOR_GRAY2BGR)
        hor = np.hstack(imgArray)
        ver = hor
    return ver

def StackedImagesV2(scale, imgList, cols):
    """
    Combine multiple images into a single display within a single window
    :param scale: The scaling factor for the images, where a value greater than 1 indicates enlargement and a value less than 1 indicates reduction.
    :param imgList: A list of images to be combined.
    :param cols: The number of images to display per row.
    :return: The combined image.
    """
    totalImages = len(imgList)
    rows = totalImages // cols if totalImages // cols * cols == totalImages else totalImages // cols + 1
    blankImages = cols * rows - totalImages

    width = imgList[0].shape[1]
    height = imgList[0].shape[0]
    imgBlank = np.zeros((height, width, 3), np.uint8)
    imgList.extend([imgBlank] * blankImages)
    for i in range(cols * rows):
        imgList[i] = cv2.resize(imgList[i], (0, 0), None, scale, scale)
        if len(imgList[i].shape) == 2:
            imgList[i] = cv2.cvtColor(imgList[i], cv2.COLOR_GRAY2BGR)
    hor = [imgBlank] * rows
    for y in range(rows):
        line = []
        for x in range(cols):
            line.append(imgList[y * cols + x])
        hor[y] = np.hstack(line)
    ver = np.vstack(hor)
    return ver

def imattributes(image):
    """Retrieve image attributes"""
    if is_numpy(image):
        imshape = image.shape
        size = imshape[:2]
        dtype = image.dtype
        height, width = size

        if len(imshape) == 2:
            depth = image.itemsize * 8
        else:
            depth = image.itemsize * 8 * imshape[2]

        return {
            "shape": imshape,
            "size": size,
            "height": height,
            "width": width,
            "dtype": dtype,
            "depth": depth,
            "source": "NumPy"
        }
    elif is_pil(image):
        width, height = image.size
        mode = image.mode

        if mode == 'L':  # Grayscale
            depth = 8
        elif mode in ('RGB', 'RGBA'):
            depth = 8 * 3 if mode == 'RGB' else 8 * 4
        elif mode in ('I', 'F'):  # Integer or float (single channel)
            depth = image.getbands()[0].size * 8
        else:
            depth = "Unknown"  # Other modes are not handled here

        return {
            "shape": (height, width, len(image.getbands())),  # Simulate a NumPy shape
            "size": (width, height),
            "height": height,
            "width": width,
            "dtype": "PIL mode " + mode,  # PIL doesn't have a direct dtype equivalent
            "depth": depth,
            "source": "PIL"
        }
    else:
        return "Input is not a NumPy array or PIL Image."

if __name__ == "__main__":
    image_path = r"E:\PythonProject\pyzjrPyPi\pyzjr\test.png"
    image = bgr_read(image_path)
    display(stacked2image(image, image))
    print(imattributes(image))


import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from pyzjr.nn.models.atten.se_module import SEAttention
from pyzjr.nn.models.bricks.conv_norm_act import ConvBNReLU, ConvBN, conv1x1


class ResBasicBlock(nn.Module):
    """The infrastructure of ResNet"""
    expansion = 1   # The output channel of a basic block is usually twice that of the input channel
    def __init__(
            self,
            in_channels,
            out_channels,
            stride=1,
            use_se=False,
            reduction=16
    ):
        super(ResBasicBlock, self).__init__()
        self.conv1 = ConvBNReLU(in_channels, out_channels, kernel_size=3, stride=stride, padding=1)
        self.conv2 = ConvBN(out_channels, out_channels, kernel_size=3, stride=1, padding=1)
        self.attention = SEAttention(out_channels, reduction)
        self.use_se = use_se
        if stride != 1 or in_channels != out_channels * self.expansion:
            self.downsample = nn.Sequential(
                conv1x1(in_channels, out_channels * self.expansion, stride),
                nn.BatchNorm2d(out_channels * self.expansion),
            )
        else:
            self.downsample = None
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.conv2(out)
        if self.use_se:
            out = self.attention(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out


class ResBottleneck(nn.Module):
    """The bottleneck structure of ResNet"""
    expansion = 4
    def __init__(
            self,
            in_channels,
            out_channels,
            stride=1,
            use_se=False,
            reduction=16
    ):
        super(ResBottleneck, self).__init__()
        groups = 1
        base_width = 64
        dilation = 1

        width = int(out_channels * (base_width / 64.)) * groups   # wide = out_channels
        self.convbnrelu1 = ConvBNReLU(in_channels, width, kernel_size=1, padding=0)  # 降维通道数
        self.convbnrelu2 = ConvBNReLU(width, width, kernel_size=3, stride=stride, dilation=dilation, groups=groups)
        self.convbn3 = ConvBN(width, out_channels * self.expansion, kernel_size=1, padding=0)   # 升维通道数
        self.attention = SEAttention(out_channels * self.expansion, reduction)
        self.relu = nn.ReLU(inplace=True)
        self.stride = stride
        self.use_se = use_se
        if stride != 1 or in_channels != out_channels * self.expansion:
            self.downsample = nn.Sequential(
                conv1x1(in_channels, out_channels * self.expansion, stride),
                nn.BatchNorm2d(out_channels * self.expansion),
            )
        else:
            self.downsample = None

    def forward(self, x):
        residual = x
        out = self.convbnrelu1(x)
        out = self.convbnrelu2(out)
        out = self.convbn3(out)
        if self.use_se:
            out = self.attention(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out

class FireModule(nn.Module):
    """SqueezeNet's compression and expansion of channels"""
    def __init__(self, in_channels, squeeze_planes, expand1x1_planes, expand3x3_planes):
        super(FireModule, self).__init__()
        self.squeeze = nn.Conv2d(in_channels, squeeze_planes, kernel_size=1)
        self.expand1x1 = nn.Conv2d(squeeze_planes, expand1x1_planes,
                                   kernel_size=1)
        self.expand3x3 = nn.Conv2d(squeeze_planes, expand3x3_planes,
                                   kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.relu(self.squeeze(x))
        return torch.cat([
            self.relu(self.expand1x1(x)),
            self.relu(self.expand3x3(x))
        ], dim=1)

class DenseBlock(nn.Module):
    """Dense connection module"""
    def __init__(
            self,
            num_layers,
            in_channels,
            bn_size,
            growth_rate,
            drop_rate=0.0
    ):
        super(DenseBlock, self).__init__()
        self.layers = nn.ModuleList()
        num_features = in_channels
        for i in range(num_layers):
            layer = self._make_dense_layer(num_features, growth_rate, bn_size, drop_rate)
            self.layers.append(layer)
            num_features += growth_rate

    def _make_dense_layer(self, num_input_features, growth_rate, bn_size, drop_rate):
        layers = [
            nn.BatchNorm2d(num_input_features),
            nn.ReLU(inplace=True),
            nn.Conv2d(num_input_features, bn_size * growth_rate, kernel_size=1, stride=1, bias=False),
            nn.BatchNorm2d(bn_size * growth_rate),
            nn.ReLU(inplace=True),
            nn.Conv2d(bn_size * growth_rate, growth_rate, kernel_size=3, stride=1, padding=1, bias=False)
        ]
        if drop_rate > 0:
            layers.append(nn.Dropout(p=drop_rate, inplace=False))
        return nn.Sequential(*layers)

    def forward(self, x):
        out = x
        for layer in self.layers:
            new_features = layer(out)
            out = torch.cat([out, new_features], 1)
        return out

class MobileInvertedResidual(nn.Module):
    """The Reverse Residual Structure of MobileNetV2"""
    def __init__(
            self,
            in_channels,
            out_channels,
            stride,
            expand_ratio
    ):
        super(MobileInvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]
        hidden_dim = int(round(in_channels * expand_ratio))
        self.use_res_connect = self.stride == 1 and in_channels == out_channels

        layers = []
        if expand_ratio != 1:
            layers.append(ConvBNReLU(in_channels, hidden_dim, kernel_size=1, padding=0))
        layers.extend([
            ConvBNReLU(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim,
                       padding=1 if stride == 1 else 0, act_layer=nn.ReLU6(inplace=True)),
            ConvBN(hidden_dim, out_channels, kernel_size=1, stride=1, padding=0),
        ])
        self.conv = nn.Sequential(*layers)
        self.out_channels = out_channels

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)

class GhostModule(nn.Module):
    """Cheap operation module for GhostNet"""
    def __init__(
            self,
            in_channels,
            out_channels,
            use_dfc=True,
            kernel_size=1,
            dw_kernel_size=3,
            ratio=2,
            stride=1,
            relu=True,
    ):
        super(GhostModule, self).__init__()
        self.use_dfc = use_dfc
        self.gate_fn = nn.Sigmoid()
        self.oup = out_channels
        init_channels = math.ceil(out_channels / ratio)
        new_channels = init_channels*(ratio-1)
        self.primary_conv = nn.Sequential(
            ConvBN(in_channels, init_channels, kernel_size, stride, kernel_size//2, bias=False),
            nn.ReLU(inplace=True) if relu else nn.Sequential(),
        )
        self.cheap_operation = nn.Sequential(
            ConvBN(init_channels, new_channels, dw_kernel_size, 1, dw_kernel_size//2, groups=init_channels, bias=False),
            nn.ReLU(inplace=True) if relu else nn.Sequential(),
        )
        self.dfc = nn.Sequential(
            # horizontal DFC and vertical DFC
            nn.Conv2d(in_channels, out_channels, kernel_size, stride, kernel_size//2, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.Conv2d(out_channels, out_channels, kernel_size=(1, 5), stride=1, padding=(0,2), groups=out_channels,bias=False),
            nn.BatchNorm2d(out_channels),
            nn.Conv2d(out_channels, out_channels, kernel_size=(5, 1), stride=1, padding=(2,0), groups=out_channels, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.AvgPool2d(kernel_size=2, stride=2)
        )

    def forward(self, x):
        x1 = self.primary_conv(x)
        x2 = self.cheap_operation(x1)
        out = torch.cat([x1, x2], dim=1)
        if self.use_dfc:
            res = self.dfc(x)
            return out[:, :self.oup, :, :] * F.interpolate(self.gate_fn(res), size=(out.shape[-2], out.shape[-1]), mode='nearest')
        else:
            return out[:, :self.oup, :, :]

class GhostBottleneck(nn.Module):
    """The Cheap Operation Bottleneck Structure of GhostNet"""
    def __init__(
            self,
            in_channels,
            mid_channels,
            out_channels,
            dw_kernel_size=3,
            stride=1,
            layer_id=None
    ):
        super(GhostBottleneck, self).__init__()
        self.stride = stride
        # Point-wise expansion
        if layer_id <= 1:
            self.ghost1 = GhostModule(in_channels, mid_channels, use_dfc=False)
        else:
            self.ghost1 = GhostModule(in_channels, mid_channels, use_dfc=True)
        # Depth-wise convolution
        if self.stride > 1:
            self.conv_bn_dw = nn.Sequential(
                nn.Conv2d(mid_channels, mid_channels, dw_kernel_size, stride=stride,
                    padding=(dw_kernel_size-1)//2, groups=mid_channels, bias=False),
                nn.BatchNorm2d(mid_channels)
            )
        self.ghost2 = GhostModule(mid_channels, out_channels, relu=False, use_dfc=False)

        # shortcut
        if (in_channels == out_channels and self.stride == 1):
            self.shortcut = nn.Sequential()
        else:
            self.shortcut = nn.Sequential(
                ConvBN(in_channels, out_channels, dw_kernel_size, stride=stride,
                       padding=(dw_kernel_size-1)//2, groups=in_channels, bias=False),
                ConvBN(in_channels, out_channels, 1, stride=1, padding=0, bias=False)
            )

    def forward(self, x):
        residual = x
        x = self.ghost1(x)
        if self.stride > 1:
            x = self.conv_bn_dw(x)
        x = self.ghost2(x)
        x += self.shortcut(residual)
        return x

if __name__=="__main__":
    input = torch.randn(32, 64, 56, 56)
    block = GhostBottleneck(in_channels=64, mid_channels=128, out_channels=64, layer_id=0)
    # block = MobileInvertedResidual(in_channels=64, out_channels=128, stride=1, expand_ratio=6)
    # block = DenseBlock(num_layers=4, in_channels=64, bn_size=4, growth_rate=12, drop_rate=0.0)
    # block = FireModule(64, squeeze_planes=16, expand1x1_planes=64, expand3x3_planes=64)
    # block = ResBottleneck(in_channels=64, out_channels=64, stride=1, use_se=True)
    output = block(input)
    print(input.shape, output.shape)


import cv2
import torch
import numpy as np
import warnings
from functools import partial
import torch.nn.functional as F
from skimage.metrics import peak_signal_noise_ratio, structural_similarity
from pyzjr.utils.check import is_numpy, is_tensor, is_gray_image
from math import exp


def gradient_metric(image, size=11, edge_pixels=1, grad_type='sobel', filter_type='blur', metric_type='max'):
    """计算指示图像中模糊强度的度量, 越接近 0 越清晰, 越接近 1 越模糊
    cv2.CV_32F : 5
    :param image: 应为灰度图
    :param size: 滤波器的大小
    :param edge_pixels: 从图像边缘裁剪的像素数量，以减少边界效应
    :param grad_type: 选择使用的梯度算子，'sobel' 或 'scharr'
    :param filter_type: 选择使用的滤波器类型，'blur', 'gaussian', 'median'
    :param metric_type: 选择返回“最大”还是“平均”的模糊度量值，'max' 或 'mean'
    :return: 图像模糊强度的度量值, 属于 (0, 1)
    """
    grad_dict = {
        'sobel': partial(cv2.Sobel, ksize=3),
        'scharr': cv2.Scharr
    }
    filter_dict = {
        'blur': partial(cv2.blur, ksize=(size, size)),
        'gaussian': partial(cv2.GaussianBlur, ksize=(size, size), sigmaX=0.3 * ((size - 1) * 0.5 - 1) + 0.8),
        'median': partial(cv2.medianBlur, ksize=size)
    }
    if grad_type not in grad_dict:
        raise ValueError(f"Invalid grad_type '{grad_type}', must be 'sobel' or 'scharr'.")
    if filter_type not in filter_dict:
        raise ValueError(f"Invalid filter_type '{filter_type}', must be 'blur', 'gaussian', or 'median'.")
    if metric_type not in ['max', 'mean']:
        raise ValueError(f"Invalid metric_type '{metric_type}', must be 'max' or 'mean'.")
    grad_operator = grad_dict[grad_type]
    filter_operator = filter_dict[filter_type]
    if not is_gray_image(image):
        warnings.warn("The input image is not a grayscale image, it will be converted to grayscale.", UserWarning)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    blur_metric = []
    h, w = image.shape[:2]
    image = image[edge_pixels:h - edge_pixels, edge_pixels:w - edge_pixels]
    blurred_image = filter_operator(image)

    for ax in range(image.ndim):  # x, y 方向
        original_gradient = np.abs(grad_operator(image, 5, ax, 1 - ax))
        blurred_gradient = np.abs(grad_operator(blurred_image, 5, ax, 1 - ax))
        g_differ = np.maximum(original_gradient - blurred_gradient, 0)
        g1 = np.sum(original_gradient)
        g2 = np.sum(g_differ)
        results = np.clip((g1 - g2) / g1, 0, 1)
        blur_metric.append(results)
    if metric_type == 'max':
        return np.max(blur_metric)
    elif metric_type == 'mean':
        return np.mean(blur_metric)

def laplacian_metric(img):
    """
    计算图像的拉普拉斯变换的方差作为模糊度量。
    拉普拉斯变换是一种边缘检测算子，对图像进行二阶微分。
    清晰的图像在边缘处会有较大的二阶微分值，而模糊的图像边缘会变得平滑，
    导致二阶微分值减小。因此，拉普拉斯变换的方差可以作为图像模糊程度的一个度量。
    方差越大，表示图像越清晰；方差越小，表示图像越模糊。

    :param img: 输入图像，应为灰度图
    :return: 拉普拉斯变换的方差值，作为模糊度量
    """
    return cv2.Laplacian(img, -1).var()

def calculate_psnrV1(input_, target_):
    """计算两张图片的PSNR"""
    img1 = np.array(input_)
    img2 = np.array(target_)
    psnr = peak_signal_noise_ratio(img1, img2)
    return psnr

def calculate_psnrV2(input_, target_):
    psnr = 0
    if is_numpy(input_) and is_numpy(target_):
        mse = np.mean((input_ - target_) ** 2)
        if mse == 0:
            return float('inf')
        max_pixel = 1. if input_.max() <= 1 else 255.
        psnr = 10 * np.log10((max_pixel ** 2) / mse)
    elif is_tensor(input_) and is_tensor(target_):
        psnr = 10 * torch.log10(1 / F.mse_loss(input_, target_)).item()
    return psnr

def calculate_ssimV1(input_, target_, win_size=7):
    """计算两张图片的 SSIM"""
    input_img = np.array(input_)
    target_img = np.array(target_)
    height, width = input_img.shape[:2]
    win_size = min(win_size, min(height, width))
    win_size = win_size + 1 if win_size % 2 == 0 else win_size
    ssim_value = structural_similarity(input_img, target_img, win_size=win_size, channel_axis=-1)
    return ssim_value

def calculate_ssimV2(input_, target_, ksize=11, sigma=1.5):
    ssim_map = 0
    if is_numpy(input_) and is_numpy(target_):
        C1 = (0.01 * 255)**2
        C2 = (0.03 * 255)**2

        img1 = input_.astype(np.float64)
        img2 = target_.astype(np.float64)
        kernel = cv2.getGaussianKernel(ksize, sigma)
        window = np.outer(kernel, kernel.transpose())
        mu1 = cv2.filter2D(img1, -1, window)[5:-5, 5:-5]
        mu2 = cv2.filter2D(img2, -1, window)[5:-5, 5:-5]
        mu1_sq = mu1**2
        mu2_sq = mu2**2
        mu1_mu2 = mu1 * mu2
        sigma1_sq = cv2.filter2D(img1**2, -1, window)[5:-5, 5:-5] - mu1_sq
        sigma2_sq = cv2.filter2D(img2**2, -1, window)[5:-5, 5:-5] - mu2_sq
        sigma12 = cv2.filter2D(img1 * img2, -1, window)[5:-5, 5:-5] - mu1_mu2

        ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) *
                                                            (sigma1_sq + sigma2_sq + C2))

    if is_tensor(input_) and is_tensor(target_):
        (_, channel, _, _) = input_.size()
        gauss = torch.Tensor([exp(-(x - ksize/2)**2/float(2 * sigma ** 2)) for x in range(ksize)])
        gauss = gauss / gauss.sum()
        _1D_window = gauss.unsqueeze(1).to(input_.device)
        _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0)
        window = _2D_window.expand(channel, 1, ksize, ksize)
        mu1 = F.conv2d(input_, window, padding = ksize//2, groups = channel)
        mu2 = F.conv2d(target_, window, padding = ksize//2, groups = channel)
        mu1_sq = mu1.pow(2)
        mu2_sq = mu2.pow(2)
        mu1_mu2 = mu1*mu2
        sigma1_sq = F.conv2d(input_*input_, window, padding=ksize//2, groups=channel) - mu1_sq
        sigma2_sq = F.conv2d(target_*target_, window, padding=ksize//2, groups=channel) - mu2_sq
        sigma12 = F.conv2d(input_*target_, window, padding=ksize//2, groups=channel) - mu1_mu2
        C1 = 0.01**2
        C2 = 0.03**2
        ssim_map = ((2*mu1_mu2 + C1)*(2*sigma12 + C2))/((mu1_sq + mu2_sq + C1)*(sigma1_sq + sigma2_sq + C2))

    return ssim_map.mean()


if __name__=="__main__":
    img1 = cv2.imread('1.png')  # input
    img2 = cv2.imread('1_1.png')  # target

    if img1.shape != img2.shape:
        print("Images must have the same dimensions.")
        exit()

    psnr_score = calculate_psnrV1(img1, img2)
    print(f'PSNR: {psnr_score:.2f} dB')
    psnr_score = calculate_psnrV2(img1, img2)
    print(f'PSNR V2: {psnr_score:.2f} dB')

    ssim_score = calculate_ssimV1(img1, img2)
    print(f'SSIM: {ssim_score:.4f}')
    ssim_score = calculate_ssimV2(img1, img2)
    print(f'SSIM V2: {ssim_score:.4f}')
    img1_tensor = torch.tensor(np.transpose(img1 / 255.0, (2, 0, 1)), dtype=torch.float32).unsqueeze(0)
    img2_tensor = torch.tensor(np.transpose(img2 / 255.0, (2, 0, 1)), dtype=torch.float32).unsqueeze(0)
    # print(img1_tensor.shape, img2_tensor.shape)
    psnr = calculate_psnrV2(img1_tensor, img2_tensor)
    print(f'PSNR V2 Tensor: {psnr:.2f} dB')
    ssim = calculate_ssimV2(img1_tensor, img2_tensor)
    print(f'SSIM V2 Tensor: {ssim:.4f}')


from .transforms import (
    Images, random_apply, random_order, random_choice, uniform_augment, Compose, RandomApply,
    RandomChoice, RandomOrder, UniformAugment, RandomHorizontalFlip, RandomVerticalFlip,
    AdjustBrightness, RandomAdjustBrightness, AdjustGamma, RandomAdjustGamma, ToHSV,
    RandomRotation, GaussianBlur, Grayscale, EqualizeHistogram, RandomEqualizeHistogram,
    Centerzoom, Resize, RandomCrop, CenterCrop, RandomResizeCrop, Pad, ResizePad, InvertColor,
    RandomInvertColor, AdjustSharpness, RandomAdjustSharpness, AdjustSaturation, RandomAdjustSaturation,
    ColorJitter, ToTensor, MeanStdNormalize, tvisionToTensor, tvisionNormalize, tvisionCenterCrop,
    tvisionRandomCrop, tvisionRandomResize, tvisionRandomContrast, tvisionRandomBrightness,
    tvisionRandomVerticalFlip, tvisionRandomHorizontalFlip, imagenet_denormal, imagenet_normal,
    min_max_normal, z_score_normal, linear_normal, zero_centered_normal, Normalizer, denormalize
)

from .augments import (
    crop_image_by_2points, crop_image_by_1points, center_crop, five_crop, centerzoom, flip,
    vertical_flip, horizontal_flip, resize, resizepad, croppad_resize, resize_samescale, translate,
    adjust_brightness_cv2, adjust_brightness_numpy, adjust_brightness_contrast, rotate,
    rotate_bound, adjust_gamma, pad_margin, erase, enhance_hsv, hist_equalize, random_lighting,
    random_rotation, random_rot90, random_horizontal_flip, random_vertical_flip, random_crop,
    random_resize_crop, addnoisy, addfog, addfog_channels, Retinex
)

from .blur import (
    meanblur, medianblur, gaussianblur, bilateralblur,
    medianfilter, meanfilter, gaussian_kernel, gaussianfilter, bilateralfilter
)

from .mask import binarization